from __future__ import annotations

import abc
import logging
from copy import deepcopy
from enum import Enum
from functools import lru_cache
from typing import Any, Callable, Dict, List, Optional, Set, Type

import structlog
from azure.core.credentials import TokenProvider
from azure.identity import (
    CertificateCredential,
    ClientSecretCredential,
    DefaultAzureCredential,
)
from pydantic import BaseModel, Field, SecretStr

from rasa.shared.providers._configs.oauth_config import OAUTH_TYPE_FIELD, OAuth

AZURE_CLIENT_ID_FIELD = "client_id"
AZURE_CLIENT_SECRET_FIELD = "client_secret"
AZURE_TENANT_ID_FIELD = "tenant_id"
AZURE_CERTIFICATE_PATH_FIELD = "certificate_path"
AZURE_CERTIFICATE_PASSWORD_FIELD = "certificate_password"
AZURE_SEND_CERTIFICATE_CHAIN_FIELD = "send_certificate_chain"
AZURE_SCOPES_FIELD = "scopes"
AZURE_AUTHORITY_FIELD = "authority_host"
AZURE_DISABLE_INSTANCE_DISCOVERY_FIELD = "disable_instance_discovery"


azure_logger = logging.getLogger("azure")
azure_logger.setLevel(logging.DEBUG)

structlogger = structlog.get_logger()


class AzureEntraIDOAuthType(str, Enum):
    """Azure Entra ID OAuth types."""

    AZURE_ENTRA_ID_DEFAULT = "azure_entra_id_default"
    AZURE_ENTRA_ID_CLIENT_SECRET = "azure_entra_id_client_secret"
    AZURE_ENTRA_ID_CLIENT_CERTIFICATE = "azure_entra_id_client_certificate"

    # Invalid type is used to indicate that the type
    # configuration is invalid EntraID or not set.
    INVALID = "invalid"

    @staticmethod
    def from_string(value: Optional[str]) -> AzureEntraIDOAuthType:
        """Converts a string to an AzureOAuthType."""
        if value is None or value not in AzureEntraIDOAuthType.valid_string_values():
            return AzureEntraIDOAuthType.INVALID

        return AzureEntraIDOAuthType(value)

    @staticmethod
    def valid_string_values() -> Set[str]:
        """Returns the valid string values for the AzureOAuthType."""
        return {e.value for e in AzureEntraIDOAuthType.valid_values()}

    @staticmethod
    def valid_values() -> Set[AzureEntraIDOAuthType]:
        """Returns the valid values for the AzureOAuthType."""
        return {
            AzureEntraIDOAuthType.AZURE_ENTRA_ID_DEFAULT,
            AzureEntraIDOAuthType.AZURE_ENTRA_ID_CLIENT_SECRET,
            AzureEntraIDOAuthType.AZURE_ENTRA_ID_CLIENT_CERTIFICATE,
        }


# BearerTokenProvider is a callable that returns a bearer token.
BearerTokenProvider = Callable[[], str]


class AzureEntraIDTokenProviderConfig(abc.ABC):
    """Interface for Azure Entra ID OAuth credential configuration."""

    @abc.abstractmethod
    def create_azure_token_provider(self) -> TokenProvider:
        """Create an Azure Entra ID token provider."""
        ...

    @classmethod
    @abc.abstractmethod
    def from_dict(
        cls: Type[AzureEntraIDTokenProviderConfig], config: Dict[str, Any]
    ) -> AzureEntraIDTokenProviderConfig:
        """Initializes a dataclass from the passed config.

        Args:
            config: (dict) The config from which to initialize.

        Returns:
            AzureEntraIDCredential
        """
        ...


class AzureEntraIDClientCredentialsConfig(AzureEntraIDTokenProviderConfig, BaseModel):
    """Azure Entra ID OAuth client credentials configuration.

    Attributes:
        client_id: The client ID.
        client_secret: The client secret.
        tenant_id: The tenant ID.
        authority_host: The authority host.
        disable_instance_discovery: Whether to disable instance discovery. This is used
            to disable fetching metadata from the Azure Instance Metadata Service.
    """

    client_id: str = Field(min_length=1)
    client_secret: SecretStr = Field(min_length=1)
    tenant_id: str = Field(min_length=1)
    authority_host: Optional[str] = None
    disable_instance_discovery: bool = False

    @staticmethod
    def required_fields() -> Set[str]:
        """Returns the required fields for the configuration."""
        return {AZURE_CLIENT_ID_FIELD, AZURE_TENANT_ID_FIELD, AZURE_CLIENT_SECRET_FIELD}

    @staticmethod
    def config_has_required_fields(config: Dict[str, Any]) -> bool:
        """Check if the configuration has all the required fields."""
        return AzureEntraIDClientCredentialsConfig.required_fields().issubset(
            set(config.keys())
        )

    @classmethod
    def from_dict(cls, config: Dict[str, Any]) -> AzureEntraIDClientCredentialsConfig:
        """Initializes a dataclass from the passed config.

        Args:
            config: (dict) The config from which to initialize.

        Returns:
            AzureClientCredentialsConfig
        """
        if not cls.config_has_required_fields(config):
            message = (
                f"A configuration for Azure client credentials "
                f"must contain the following keys: {cls.required_fields()}"
            )
            structlogger.error(
                "azure_client_credentials_config.missing_required_keys",
                message=message,
                config=config,
            )
            raise ValueError(message)

        return cls(
            client_id=config.pop(AZURE_CLIENT_ID_FIELD),
            client_secret=config.pop(AZURE_CLIENT_SECRET_FIELD),
            tenant_id=config.pop(AZURE_TENANT_ID_FIELD),
            authority_host=config.pop(AZURE_AUTHORITY_FIELD, None),
            disable_instance_discovery=config.pop(
                AZURE_DISABLE_INSTANCE_DISCOVERY_FIELD, False
            ),
        )

    def create_azure_token_provider(self) -> TokenProvider:
        """Create a ClientSecretCredential for Azure Entra ID."""
        return create_azure_entra_id_client_credentials(
            client_id=self.client_id,
            client_secret=self.client_secret.get_secret_value(),
            tenant_id=self.tenant_id,
            authority_host=self.authority_host,
            disable_instance_discovery=self.disable_instance_discovery,
        )


# We are caching the result of this function to preserve the refresh
# token which is stored inside the credential object.
# This allows us to reuse the same credential object (refresh token)
# across multiple requests.
# Refresh token is used to get a new access token when the current access
# token expires without having to re-authenticate the
# user (transmit the client secret again).
@lru_cache
def create_azure_entra_id_client_credentials(
    client_id: str,
    client_secret: str,
    tenant_id: str,
    authority_host: Optional[str] = None,
    disable_instance_discovery: bool = False,
) -> ClientSecretCredential:
    """Creates a ClientSecretCredential for Azure Entra ID.

    We cache the result of this function to avoid creating multiple instances
    of the same credential. This makes it possible to utilise the token caching
    and token refreshing functionality of the azure-identity library.

    Args:
        client_id: The client ID.
        client_secret: The client secret.
        tenant_id: The tenant ID.
        authority_host: The authority host.
        disable_instance_discovery: Whether to disable instance discovery. This is used
            to disable fetching metadata from the Azure Instance Metadata Service.

    Returns:
        ClientSecretCredential
    """
    return ClientSecretCredential(
        client_id=client_id,
        client_secret=client_secret,
        tenant_id=tenant_id,
        authority=authority_host,
        disable_instance_discovery=disable_instance_discovery,
    )


class AzureEntraIDClientCertificateConfig(AzureEntraIDTokenProviderConfig, BaseModel):
    """Azure Entra ID OAuth client certificate configuration.

    Attributes:
        client_id: The client ID.
        tenant_id: The tenant ID.
        certificate_path: The path to the certificate file.
        certificate_password: The certificate password.
        send_certificate_chain: Whether to send the certificate chain.
        authority_host: The authority host.
        disable_instance_discovery: Whether to disable instance discovery. This is used
            to disable fetching metadata from the Azure Instance Metadata Service.
    """

    client_id: str = Field(min_length=1)
    tenant_id: str = Field(min_length=1)
    certificate_path: str = Field(min_length=1)
    certificate_password: Optional[SecretStr] = None
    send_certificate_chain: bool = False
    authority_host: Optional[str] = None
    disable_instance_discovery: bool = False

    @staticmethod
    def required_fields() -> Set[str]:
        """Returns the required fields for the configuration."""
        return {
            AZURE_CLIENT_ID_FIELD,
            AZURE_TENANT_ID_FIELD,
            AZURE_CERTIFICATE_PATH_FIELD,
        }

    @staticmethod
    def config_has_required_fields(config: Dict[str, Any]) -> bool:
        """Check if the configuration has all the required fields."""
        return AzureEntraIDClientCertificateConfig.required_fields().issubset(
            set(config.keys())
        )

    @classmethod
    def from_dict(cls, config: Dict[str, Any]) -> AzureEntraIDClientCertificateConfig:
        """Initializes a dataclass from the passed config.

        Args:
            config: (dict) The config from which to initialize.

        Returns:
            AzureClientCertificateConfig
        """
        if not cls.config_has_required_fields(config):
            message = (
                f"A configuration for Azure client certificate "
                f"must contain "
                f"the following keys: {cls.required_fields()}"
            )
            structlogger.error(
                "azure_client_certificate_config.validation_error",
                message=message,
                config=config,
            )
            raise ValueError(message)

        return cls(
            client_id=config[AZURE_CLIENT_ID_FIELD],
            tenant_id=config[AZURE_TENANT_ID_FIELD],
            certificate_path=config[AZURE_CERTIFICATE_PATH_FIELD],
            certificate_password=config.get(AZURE_CERTIFICATE_PASSWORD_FIELD, None),
            authority_host=config.get(AZURE_AUTHORITY_FIELD, None),
            send_certificate_chain=config.get(
                AZURE_SEND_CERTIFICATE_CHAIN_FIELD, False
            ),
            disable_instance_discovery=config.get(
                AZURE_DISABLE_INSTANCE_DISCOVERY_FIELD, False
            ),
        )

    def create_azure_token_provider(self) -> TokenProvider:
        """Creates a CertificateCredential for Azure Entra ID."""
        return create_azure_entra_id_certificate_credentials(
            client_id=self.client_id,
            tenant_id=self.tenant_id,
            certificate_path=self.certificate_path,
            password=self.certificate_password.get_secret_value()
            if self.certificate_password
            else None,
            send_certificate_chain=self.send_certificate_chain,
            authority_host=self.authority_host,
            disable_instance_discovery=self.disable_instance_discovery,
        )


# We are caching the result of this function to preserve the refresh
# token which is stored inside the credential object.
# This allows us to reuse the same credential object (refresh token)
# across multiple requests.
# Refresh token is used to get a new access token when the current
# access token expires without having to re-authenticate
# the user (transmit the client certificate again).
@lru_cache
def create_azure_entra_id_certificate_credentials(
    tenant_id: str,
    client_id: str,
    certificate_path: Optional[str] = None,
    password: Optional[str] = None,
    send_certificate_chain: bool = False,
    authority_host: Optional[str] = None,
    disable_instance_discovery: bool = False,
) -> CertificateCredential:
    """Creates a CertificateCredential for Azure Entra ID.

    We cache the result of this function to avoid creating multiple instances
    of the same credential. This makes it possible to utilise the token caching
    and token refreshing functionality of the azure-identity library.

    Args:
        tenant_id: The tenant ID.
        client_id: The client ID.
        certificate_path: The path to the certificate file.
        password: The certificate password.
        send_certificate_chain: Whether to send the certificate chain.
        authority_host: The authority host.
        disable_instance_discovery: Whether to disable instance discovery. This is used

    Returns:
        CertificateCredential
    """

    return CertificateCredential(
        client_id=client_id,
        tenant_id=tenant_id,
        certificate_path=certificate_path,
        password=password.encode("utf-8") if password else None,
        send_certificate_chain=send_certificate_chain,
        authority=authority_host,
        disable_instance_discovery=disable_instance_discovery,
    )


class AzureEntraIDDefaultCredentialsConfig(AzureEntraIDTokenProviderConfig, BaseModel):
    """Azure Entra ID OAuth default credentials configuration.

    Attributes:
        authority_host: The authority host.
    """

    authority_host: Optional[str] = None

    @classmethod
    def from_dict(cls, config: Dict[str, Any]) -> AzureEntraIDDefaultCredentialsConfig:
        """Initializes a dataclass from the passed config.

        Args:
            config: (dict) The config from which to initialize.

        Returns:
            AzureOAuthDefaultCredentialsConfig
        """
        return cls(authority_host=config.pop(AZURE_AUTHORITY_FIELD, None))

    def create_azure_token_provider(self) -> TokenProvider:
        """Creates a DefaultAzureCredential."""
        return create_azure_entra_id_default_credentials(
            authority_host=self.authority_host
        )


@lru_cache
def create_azure_entra_id_default_credentials(
    authority_host: Optional[str] = None,
) -> DefaultAzureCredential:
    """Creates a DefaultAzureCredential.

    We cache the result of this function to avoid creating multiple instances
    of the same credential. This makes it possible to utilise the token caching
    functionality of the azure-identity library.

    Args:
        authority_host: The authority host.

    Returns:
        DefaultAzureCredential
    """
    return DefaultAzureCredential(authority=authority_host)


class AzureEntraIDOAuthConfig(OAuth, BaseModel):
    """Azure Entra ID OAuth configuration.

    It consists of the scopes and the Azure Entra ID OAuth credentials.
    """

    # pydantic configuration to allow arbitrary user defined types
    class Config:
        arbitrary_types_allowed = True

    scopes: List[str]
    azure_entra_id_token_provider_config: AzureEntraIDTokenProviderConfig

    @staticmethod
    def _supported_azure_oauth() -> (
        Dict[AzureEntraIDOAuthType, Type[AzureEntraIDTokenProviderConfig]]
    ):
        """Returns a mapping of supported Azure Entra ID OAuth types to their"""
        return {
            AzureEntraIDOAuthType.AZURE_ENTRA_ID_DEFAULT: AzureEntraIDDefaultCredentialsConfig,  # noqa: E501
            AzureEntraIDOAuthType.AZURE_ENTRA_ID_CLIENT_SECRET: AzureEntraIDClientCredentialsConfig,  # noqa: E501
            AzureEntraIDOAuthType.AZURE_ENTRA_ID_CLIENT_CERTIFICATE: AzureEntraIDClientCertificateConfig,  # noqa: E501
        }

    @staticmethod
    def _get_azure_oauth_by_type(
        oauth_type: AzureEntraIDOAuthType,
    ) -> Type[AzureEntraIDTokenProviderConfig]:
        """Returns the Azure Entra ID OAuth class based on the type.

        Args:
            oauth_type: (AzureOAuthType) The type of the Azure Entra ID OAuth.

        Returns:
            The Azure Entra ID OAuth class

        Raises:
            ValueError: If the passed oauth_type is not supported or invalid.
        """
        azure_oauth_types = AzureEntraIDOAuthConfig._supported_azure_oauth()
        azure_oauth_class = azure_oauth_types.get(oauth_type)

        if azure_oauth_class is None:
            message = (
                f"Unsupported Azure Entra ID oauth type: {oauth_type}. "
                f"Supported types are: {AzureEntraIDOAuthType.valid_string_values()}"
            )
            structlogger.error(
                "azure_oauth_config.unsupported_azure_oauth_type",
                message=message,
            )
            raise ValueError(message)

        return azure_oauth_class

    @classmethod
    def from_dict(cls, oauth_config: Dict[str, Any]) -> AzureEntraIDOAuthConfig:
        """Initializes a dataclass from the passed config.

        Args:
            oauth_config: (dict) The config from which to initialize.

        Returns:
            AzureOAuthConfig
        """

        config = deepcopy(oauth_config)

        scopes = AzureEntraIDOAuthConfig._read_scopes_from_config(config)
        azure_credentials = (
            AzureEntraIDOAuthConfig._create_azure_entra_id_client_from_config(config)
        )
        return cls(
            azure_entra_id_token_provider_config=azure_credentials, scopes=scopes
        )

    @staticmethod
    def _read_scopes_from_config(oauth_config: Dict[str, Any]) -> List[str]:
        """Reads scopes from the configuration.

        The original scopes are removed from the configuration.

        Args:
            oauth_config: (dict) The configuration from which to read the scopes.

        Returns:
            List[str]: The list of scopes.
        """
        scopes = oauth_config.pop(AZURE_SCOPES_FIELD, "")

        if not scopes:
            message = "Azure Entra ID scopes cannot be empty."
            structlogger.error(
                "azure_oauth_config.scopes_empty",
                message=message,
            )
            raise ValueError(message)

        if isinstance(scopes, str):
            scopes = [scopes]

        return scopes

    @staticmethod
    def _create_azure_entra_id_client_from_config(
        oauth_config: Dict[str, Any],
    ) -> AzureEntraIDTokenProviderConfig:
        """Creates an Azure Entra ID client from the configuration.

        Args:
            oauth_config: (dict) The configuration from which to create the credential.

        Returns:
            AzureEntraIDTokenProviderConfig: The Azure OAuth credential.
        """

        oauth_type = AzureEntraIDOAuthType.from_string(
            oauth_config.pop(OAUTH_TYPE_FIELD, None)
        )

        if oauth_type == AzureEntraIDOAuthType.INVALID:
            message = (
                "Azure Entra ID oauth configuration must contain "
                f"'{OAUTH_TYPE_FIELD}' field and it must be set to one of the "
                f"following values: {AzureEntraIDOAuthType.valid_string_values()}, "
            )
            structlogger.error(
                "azure_oauth_config.missing_azure_oauth_type",
                message=message,
            )
            raise ValueError(message)

        azure_oauth_class = AzureEntraIDOAuthConfig._get_azure_oauth_by_type(oauth_type)
        return azure_oauth_class.from_dict(oauth_config)

    def _create_azure_credential(
        self,
    ) -> TokenProvider:
        """Create an Azure Entra ID client which can be used to get a bearer token."""
        return self.azure_entra_id_token_provider_config.create_azure_token_provider()

    def get_bearer_token(self) -> str:
        """Returns a bearer token."""
        return self._create_azure_credential().get_token(*self.scopes).token  # type: ignore



import argparse
import logging
from typing import TYPE_CHECKING, List, Optional, Text, Union

import pluggy

# IMPORTANT: do not import anything from rasa here - use scoped imports
#  this avoids circular imports, as the hooks are used in different places
#  across the codebase.

if TYPE_CHECKING:
    from rasa.anonymization.anonymization_pipeline import AnonymizationPipeline
    from rasa.cli import SubParsersAction
    from rasa.core.brokers.broker import EventBroker
    from rasa.core.tracker_store import TrackerStore
    from rasa.shared.core.domain import Domain
    from rasa.utils.endpoints import EndpointConfig

hookimpl = pluggy.HookimplMarker("rasa")
logger = logging.getLogger(__name__)


@hookimpl  # type: ignore[misc]
def refine_cli(
    subparsers: "SubParsersAction",
    parent_parsers: List[argparse.ArgumentParser],
) -> None:
    from rasa.cli import dialogue_understanding_test, e2e_test, inspect, markers
    from rasa.cli import license as license_cli
    from rasa.cli.studio import studio

    e2e_test.add_subparser(subparsers, parent_parsers)
    dialogue_understanding_test.add_subparser(subparsers, parent_parsers)
    studio.add_subparser(subparsers, parent_parsers)
    license_cli.add_subparser(subparsers, parent_parsers)
    markers.add_subparser(subparsers, parent_parsers)
    inspect.add_subparser(subparsers, parent_parsers)
    return None


@hookimpl  # type: ignore[misc]
def configure_commandline(cmdline_arguments: argparse.Namespace) -> Optional[Text]:
    from rasa.cli import x as rasa_x
    from rasa.tracing import config

    endpoints_file = None

    if cmdline_arguments.func.__name__ == "rasa_x":
        _, endpoints_file = rasa_x._get_credentials_and_endpoints_paths(
            cmdline_arguments
        )
    elif "endpoints" in cmdline_arguments:
        endpoints_file = cmdline_arguments.endpoints

    if endpoints_file is not None:
        tracer_provider = config.get_tracer_provider(endpoints_file)
        config.configure_tracing(tracer_provider)
        config.configure_metrics(endpoints_file)

    return endpoints_file


@hookimpl  # type: ignore[misc]
def init_telemetry(endpoints_file: Optional[Text]) -> None:
    import rasa.telemetry

    rasa.telemetry.identify_endpoint_config_traits(endpoints_file)


@hookimpl  # type: ignore[misc]
def init_managers(endpoints_file: Optional[Text]) -> None:
    from rasa.core.secrets_manager.factory import load_secret_manager

    load_secret_manager(endpoints_file)


@hookimpl  # type: ignore[misc]
def create_tracker_store(
    endpoint_config: Union["TrackerStore", "EndpointConfig"],
    domain: "Domain",
    event_broker: Optional["EventBroker"],
) -> "TrackerStore":
    from rasa.core.auth_retry_tracker_store import AuthRetryTrackerStore
    from rasa.utils.endpoints import EndpointConfig

    if isinstance(endpoint_config, EndpointConfig):
        return AuthRetryTrackerStore(
            endpoint_config=endpoint_config, domain=domain, event_broker=event_broker
        )
    return endpoint_config


@hookimpl  # type: ignore[misc]
def init_anonymization_pipeline(endpoints_file: Optional[Text]) -> None:
    """Hook implementation for initializing the anonymization pipeline."""
    from rasa.anonymization.anonymization_pipeline import load_anonymization_pipeline

    load_anonymization_pipeline(endpoints_file)


@hookimpl  # type: ignore[misc]
def get_anonymization_pipeline() -> Optional["AnonymizationPipeline"]:
    """Hook implementation for getting the anonymization pipeline."""
    from rasa.anonymization.anonymization_pipeline import AnonymizationPipelineProvider

    return AnonymizationPipelineProvider().get_anonymization_pipeline()


@hookimpl  # type: ignore[misc]
def after_server_stop() -> None:
    """Hook implementation for stopping the anonymization pipeline."""
    from rasa.anonymization.anonymization_pipeline import AnonymizationPipelineProvider

    anon_pipeline = AnonymizationPipelineProvider().get_anonymization_pipeline()

    if anon_pipeline is not None:
        anon_pipeline.stop()



import importlib.resources
from typing import Any, Dict, Optional, Text

import structlog

from rasa.dialogue_understanding.commands.command_syntax_manager import (
    CommandSyntaxManager,
    CommandSyntaxVersion,
)
from rasa.dialogue_understanding.generator.constants import (
    FLOW_RETRIEVAL_KEY,
    LLM_CONFIG_KEY,
    USER_INPUT_CONFIG_KEY,
)
from rasa.dialogue_understanding.generator.flow_retrieval import FlowRetrieval
from rasa.dialogue_understanding.generator.single_step.compact_llm_command_generator import (  # noqa: E501
    CompactLLMCommandGenerator,
)
from rasa.engine.recipes.default_recipe import DefaultV1Recipe
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.shared.constants import (
    EMBEDDINGS_CONFIG_KEY,
    PROMPT_CONFIG_KEY,
    PROMPT_TEMPLATE_CONFIG_KEY,
)
from rasa.shared.utils.io import deep_container_fingerprint
from rasa.shared.utils.llm import get_prompt_template, resolve_model_client_config

DEFAULT_COMMAND_PROMPT_TEMPLATE = importlib.resources.read_text(
    "rasa.dialogue_understanding.generator.prompt_templates",
    "command_prompt_template.jinja2",
)


structlogger = structlog.get_logger()


@DefaultV1Recipe.register(
    [
        DefaultV1Recipe.ComponentType.COMMAND_GENERATOR,
    ],
    is_trainable=True,
)
class SingleStepLLMCommandGenerator(CompactLLMCommandGenerator):
    """A single step LLM-based command generator."""

    def __init__(
        self,
        config: Dict[str, Any],
        model_storage: ModelStorage,
        resource: Resource,
        prompt_template: Optional[Text] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            config,
            model_storage,
            resource,
            prompt_template=prompt_template,
            **kwargs,
        )

        # Set the prompt template
        if config.get(PROMPT_CONFIG_KEY):
            structlogger.warning(
                "single_step_llm_command_generator.init",
                event_info=(
                    "The config parameter 'prompt' is deprecated "
                    "and will be removed in Rasa 4.0.0. "
                    "Please use the config parameter 'prompt_template' instead. "
                ),
            )
        config_prompt = (
            config.get(PROMPT_CONFIG_KEY)
            or config.get(PROMPT_TEMPLATE_CONFIG_KEY)
            or None
        )
        self.prompt_template = prompt_template or get_prompt_template(
            config_prompt,
            DEFAULT_COMMAND_PROMPT_TEMPLATE,
        )

        # Set the command syntax version to v1
        CommandSyntaxManager.set_syntax_version(CommandSyntaxVersion.v1)

    @staticmethod
    def get_default_config() -> Dict[str, Any]:
        """The component's default config (see parent class for full docstring)."""
        return {
            PROMPT_CONFIG_KEY: None,  # Legacy
            PROMPT_TEMPLATE_CONFIG_KEY: None,
            USER_INPUT_CONFIG_KEY: None,
            LLM_CONFIG_KEY: None,
            FLOW_RETRIEVAL_KEY: FlowRetrieval.get_default_config(),
        }

    @classmethod
    def fingerprint_addon(cls: Any, config: Dict[str, Any]) -> Optional[str]:
        """Add a fingerprint for the graph."""
        config_prompt = (
            config.get(PROMPT_CONFIG_KEY)
            or config.get(PROMPT_TEMPLATE_CONFIG_KEY)
            or None
        )
        prompt_template = get_prompt_template(
            config_prompt,
            DEFAULT_COMMAND_PROMPT_TEMPLATE,
        )
        llm_config = resolve_model_client_config(
            config.get(LLM_CONFIG_KEY), SingleStepLLMCommandGenerator.__name__
        )
        embedding_config = resolve_model_client_config(
            config.get(FLOW_RETRIEVAL_KEY, {}).get(EMBEDDINGS_CONFIG_KEY),
            FlowRetrieval.__name__,
        )
        return deep_container_fingerprint(
            [prompt_template, llm_config, embedding_config]
        )



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Any, Dict, List

from rasa.dialogue_understanding.commands.command import Command
from rasa.dialogue_understanding.commands.command_syntax_manager import (
    CommandSyntaxManager,
    CommandSyntaxVersion,
)
from rasa.dialogue_understanding.patterns.repeat import (
    RepeatBotMessagesPatternFlowStackFrame,
)
from rasa.shared.core.events import Event
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.trackers import DialogueStateTracker


@dataclass
class RepeatBotMessagesCommand(Command):
    """A command to indicate that the bot should repeat its last messages."""

    @classmethod
    def command(cls) -> str:
        """Returns the command type."""
        return "repeat"

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> RepeatBotMessagesCommand:
        """Converts the dictionary to a command.

        Returns:
            The converted dictionary.
        """
        return RepeatBotMessagesCommand()

    def run_command_on_tracker(
        self,
        tracker: DialogueStateTracker,
        all_flows: FlowsList,
        original_tracker: DialogueStateTracker,
    ) -> List[Event]:
        """Runs the command on the tracker.
        Get all the bot utterances until last user utterance and repeat them.

        Args:
            tracker: The tracker to run the command on.
            all_flows: All flows in the assistant.
            original_tracker: The tracker before any command was executed.

        Returns:
            The events to apply to the tracker.
        """
        stack = tracker.stack
        stack.push(RepeatBotMessagesPatternFlowStackFrame())
        return tracker.create_stack_updated_events(stack)

    def __hash__(self) -> int:
        return hash(self.command())

    def __eq__(self, other: object) -> bool:
        return isinstance(other, RepeatBotMessagesCommand)

    def to_dsl(self) -> str:
        """Converts the command to a DSL string."""
        mapper = {
            CommandSyntaxVersion.v1: "RepeatLastBotMessages()",
            CommandSyntaxVersion.v2: "repeat message",
        }
        return mapper.get(
            CommandSyntaxManager.get_syntax_version(),
            mapper[CommandSyntaxManager.get_default_syntax_version()],
        )

    @classmethod
    def from_dsl(cls, match: re.Match, **kwargs: Any) -> RepeatBotMessagesCommand:
        """Converts the DSL string to a command."""
        return RepeatBotMessagesCommand()

    @staticmethod
    def regex_pattern() -> str:
        mapper = {
            CommandSyntaxVersion.v1: r"RepeatLastBotMessages\(\)",
            CommandSyntaxVersion.v2: r"^[^\w]*repeat message$",
        }
        return mapper.get(
            CommandSyntaxManager.get_syntax_version(),
            mapper[CommandSyntaxManager.get_default_syntax_version()],
        )



from __future__ import annotations

import logging
from typing import Any, Dict, List, Union

import structlog

from rasa.shared.exceptions import ProviderClientAPIException
from rasa.shared.providers._configs.litellm_router_client_config import (
    LiteLLMRouterClientConfig,
)
from rasa.shared.providers.constants import LITE_LLM_MODEL_FIELD
from rasa.shared.providers.llm._base_litellm_client import _BaseLiteLLMClient
from rasa.shared.providers.llm.llm_response import LLMResponse
from rasa.shared.providers.router._base_litellm_router_client import (
    _BaseLiteLLMRouterClient,
)
from rasa.shared.utils.io import suppress_logs

structlogger = structlog.get_logger()


class LiteLLMRouterLLMClient(_BaseLiteLLMRouterClient, _BaseLiteLLMClient):
    """A client for interfacing with LiteLLM Router LLM endpoints.

    Parameters:
        model_group_id (str): The model group ID.
        model_configurations (List[Dict[str, Any]]): The list of model configurations.
        router_settings (Dict[str, Any]): The router settings.
        kwargs (Optional[Dict[str, Any]]): Additional configuration parameters.

    Raises:
        ProviderClientValidationError: If validation of the client setup fails.
    """

    def __init__(
        self,
        model_group_id: str,
        model_configurations: List[Dict[str, Any]],
        router_settings: Dict[str, Any],
        **kwargs: Any,
    ):
        super().__init__(
            model_group_id, model_configurations, router_settings, **kwargs
        )

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> LiteLLMRouterLLMClient:
        """Instantiates a LiteLLM Router LLM client from a configuration dict.

        Args:
            config: (Dict[str, Any]) The configuration dictionary.

        Returns:
            LiteLLMRouterLLMClient: The instantiated LiteLLM Router LLM client.

        Raises:
            ValueError: If the configuration is invalid.
        """
        try:
            client_config = LiteLLMRouterClientConfig.from_dict(config)
        except (KeyError, ValueError) as e:
            message = "Cannot instantiate a client from the passed configuration."
            structlogger.error(
                "litellm_router_llm_client.from_config.error",
                message=message,
                config=config,
                original_error=e,
            )
            raise

        return cls(
            model_group_id=client_config.model_group_id,
            model_configurations=client_config.litellm_model_list,
            router_settings=client_config.litellm_router_settings,
            use_chat_completions_endpoint=client_config.use_chat_completions_endpoint,
            **client_config.extra_parameters,
        )

    @suppress_logs(log_level=logging.WARNING)
    def _text_completion(self, prompt: Union[List[str], str]) -> LLMResponse:
        """
        Synchronously generate completions for given prompt.

        Args:
            prompt: Prompt to generate the completion for.
        Returns:
            List of message completions.
        Raises:
            ProviderClientAPIException: If the API request fails.
        """
        try:
            structlogger.info(
                "litellm_router_llm_client.text_completion",
                _completion_fn_args=self._completion_fn_args,
            )
            response = self.router_client.text_completion(
                prompt=prompt, **self._completion_fn_args
            )
            return self._format_text_completion_response(response)
        except Exception as e:
            raise ProviderClientAPIException(e)

    @suppress_logs(log_level=logging.WARNING)
    async def _atext_completion(self, prompt: Union[List[str], str]) -> LLMResponse:
        """
        Asynchronously generate completions for given prompt.

        Args:
            prompt: Prompt to generate the completion for.
        Returns:
            List of message completions.
        Raises:
            ProviderClientAPIException: If the API request fails.
        """
        try:
            response = await self.router_client.atext_completion(
                prompt=prompt, **self._completion_fn_args
            )
            return self._format_text_completion_response(response)
        except Exception as e:
            raise ProviderClientAPIException(e)

    @suppress_logs(log_level=logging.WARNING)
    def completion(self, messages: Union[List[dict], List[str], str]) -> LLMResponse:
        """
        Synchronously generate completions for given list of messages.

        Method overrides the base class method to call the appropriate
        completion method based on the configuration. If the chat completions
        endpoint is enabled, the completion method is called. Otherwise, the
        text_completion method is called.

        Args:
            messages: The message can be,
                - a list of preformatted messages. Each message should be a dictionary
                    with the following keys:
                    - content: The message content.
                    - role: The role of the message (e.g. user or system).
                - a list of messages. Each message is a string and will be formatted
                    as a user message.
                - a single message as a string which will be formatted as user message.
        Returns:
            List of message completions.
        Raises:
            ProviderClientAPIException: If the API request fails.
        """
        if not self._use_chat_completions_endpoint:
            return self._text_completion(messages)
        try:
            formatted_messages = self._format_messages(messages)
            response = self.router_client.completion(
                messages=formatted_messages, **self._completion_fn_args
            )
            return self._format_response(response)
        except Exception as e:
            raise ProviderClientAPIException(e)

    @suppress_logs(log_level=logging.WARNING)
    async def acompletion(
        self, messages: Union[List[dict], List[str], str]
    ) -> LLMResponse:
        """
        Asynchronously generate completions for given list of messages.

        Method overrides the base class method to call the appropriate
        completion method based on the configuration. If the chat completions
        endpoint is enabled, the completion method is called. Otherwise, the
        text_completion method is called.

        Args:
            messages: The message can be,
                - a list of preformatted messages. Each message should be a dictionary
                    with the following keys:
                    - content: The message content.
                    - role: The role of the message (e.g. user or system).
                - a list of messages. Each message is a string and will be formatted
                    as a user message.
                - a single message as a string which will be formatted as user message.
        Returns:
            List of message completions.
        Raises:
            ProviderClientAPIException: If the API request fails.
        """
        if not self._use_chat_completions_endpoint:
            return await self._atext_completion(messages)
        try:
            formatted_messages = self._format_messages(messages)
            response = await self.router_client.acompletion(
                messages=formatted_messages, **self._completion_fn_args
            )
            return self._format_response(response)
        except Exception as e:
            raise ProviderClientAPIException(e)

    @property
    def _completion_fn_args(self) -> Dict[str, Any]:
        """Returns the completion arguments for invoking a call through
        LiteLLM's completion functions.
        """
        return {
            **self._litellm_extra_parameters,
            LITE_LLM_MODEL_FIELD: self.model_group_id,
        }



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List

import structlog

from rasa.dialogue_understanding.commands.command import Command
from rasa.dialogue_understanding.patterns.cannot_handle import (
    CannotHandlePatternFlowStackFrame,
)
from rasa.dialogue_understanding.patterns.handle_digressions import (
    HandleDigressionsPatternFlowStackFrame,
)
from rasa.dialogue_understanding.stack.utils import (
    top_flow_frame,
    user_flows_on_the_stack,
)
from rasa.shared.core.events import Event
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.flows.steps import CollectInformationFlowStep
from rasa.shared.core.flows.utils import ALL_LABEL
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.nlu.constants import HANDLE_DIGRESSIONS_COMMAND

structlogger = structlog.get_logger()


@dataclass
class HandleDigressionsCommand(Command):
    """A command to handle digressions during an active flow."""

    flow: str
    """The interrupting flow."""

    @classmethod
    def command(cls) -> str:
        """Returns the command type."""
        return HANDLE_DIGRESSIONS_COMMAND

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> HandleDigressionsCommand:
        """Converts the dictionary to a command.

        Returns:
            The converted dictionary.
        """
        try:
            return HandleDigressionsCommand(flow=data["flow"])
        except KeyError as e:
            raise ValueError(
                f"Missing parameter '{e}' while parsing HandleDigressionsCommand."
            ) from e

    def run_command_on_tracker(
        self,
        tracker: DialogueStateTracker,
        all_flows: FlowsList,
        original_tracker: DialogueStateTracker,
    ) -> List[Event]:
        """Runs the command on the tracker.

        Args:
            tracker: The tracker to run the command on.
            all_flows: All flows in the assistant.
            original_tracker: The tracker before any command was executed.

        Returns:
            The events to apply to the tracker.
        """
        stack = tracker.stack
        original_stack = original_tracker.stack

        if self.flow in user_flows_on_the_stack(stack):
            structlogger.debug(
                "command_executor.skip_command.already_started_flow", command=self
            )
            return []
        elif self.flow not in all_flows.flow_ids:
            structlogger.debug(
                "command_executor.push_cannot_handle.start_invalid_flow_id",
                command=self,
            )
            stack.push(CannotHandlePatternFlowStackFrame())
            return tracker.create_stack_updated_events(stack)

        # this allows to include called user flows in the stack search
        latest_user_frame = top_flow_frame(original_stack, ignore_call_frames=False)

        if latest_user_frame is None:
            structlogger.debug(
                "command_executor.skip_command.no_top_flow", command=self
            )
            return []

        original_top_flow = latest_user_frame.flow(all_flows)
        current_step = original_top_flow.step_by_id(latest_user_frame.step_id)
        if not isinstance(current_step, CollectInformationFlowStep):
            structlogger.debug(
                "command_executor.skip_command.not_at_a_collect_step", command=self
            )
            return []

        ask_confirm_digressions = set(
            current_step.ask_confirm_digressions
            + original_top_flow.ask_confirm_digressions
        )

        block_digressions = set(
            current_step.block_digressions + original_top_flow.block_digressions
        )

        if block_digressions:
            if ALL_LABEL in block_digressions:
                block_digressions.remove(ALL_LABEL)
                block_digressions.add(self.flow)

        if ask_confirm_digressions:
            if ALL_LABEL in ask_confirm_digressions:
                ask_confirm_digressions.remove(ALL_LABEL)
                ask_confirm_digressions.add(self.flow)

        structlogger.debug(
            "command_executor.push_handle_digressions",
            interrupting_flow_id=self.flow,
            interrupted_flow_id=original_top_flow.id,
            interrupted_step_id=current_step.id,
            ask_confirm_digressions=ask_confirm_digressions,
            block_digressions=block_digressions,
        )
        stack.push(
            HandleDigressionsPatternFlowStackFrame(
                interrupting_flow_id=self.flow,
                interrupted_flow_id=original_top_flow.id,
                interrupted_step_id=current_step.id,
                ask_confirm_digressions=ask_confirm_digressions,
                block_digressions=block_digressions,
            )
        )

        return tracker.create_stack_updated_events(stack)

    def __hash__(self) -> int:
        return hash(self.flow)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, HandleDigressionsCommand):
            return False

        return other.flow == self.flow



import copy
import typing
from typing import Any, Dict, List, Optional, Text

import numpy as np
from pydantic import BaseModel

from rasa.dialogue_understanding.commands.prompt_command import PromptCommand
from rasa.dialogue_understanding_test.du_test_case import (
    DialogueUnderstandingTestCase,
    DialogueUnderstandingTestStep,
)
from rasa.dialogue_understanding_test.utils import get_command_comparison
from rasa.shared.nlu.constants import KEY_SYSTEM_PROMPT, KEY_USER_PROMPT

if typing.TYPE_CHECKING:
    from rasa.dialogue_understanding_test.command_metric_calculation import (
        CommandMetrics,
    )

KEY_TEST_CASES_ACCURACY = "test_cases"
KEY_USER_UTTERANCES_ACCURACY = "user_utterances"

OUTPUT_NUMBER_OF_FAILED_TESTS = "number_of_failed_tests"
OUTPUT_NUMBER_OF_PASSED_TESTS = "number_of_passed_tests"
OUTPUT_TEST_CASES_ACCURACY = "test_cases_accuracy"
OUTPUT_USER_UTTERANCES_ACCURACY = "user_utterances_accuracy"
OUTPUT_NUMBER_OF_PASSED_USER_UTTERANCES = "number_of_passed_user_utterances"
OUTPUT_NUMBER_OF_FAILED_USER_UTTERANCES = "number_of_failed_user_utterances"
OUTPUT_COMMAND_METRICS = "command_metrics"
OUTPUT_LATENCY_METRICS = "latency"
OUTPUT_COMPLETION_TOKEN_METRICS = "completion_token"
OUTPUT_PROMPT_TOKEN_METRICS = "prompt_token"
OUTPUT_NAMES_OF_FAILED_TESTS = "names_of_failed_tests"
OUTPUT_NAMES_OF_PASSED_TESTS = "names_of_passed_tests"
OUTPUT_LLM_COMMAND_GENERATOR_CONFIG = "llm_command_generator_config"


class DialogueUnderstandingTestResult(BaseModel):
    """Result of a single dialogue understanding test case."""

    test_case: DialogueUnderstandingTestCase
    passed: bool
    error_line: Optional[int] = None

    def get_expected_commands(self) -> List[PromptCommand]:
        return self.test_case.get_expected_commands()


class FailedTestStep(BaseModel):
    """Failed test step information."""

    file: str
    test_case_name: str
    failed_user_utterance: str
    error_line: int
    pass_status: bool
    command_generators: List[str]
    prompts: Optional[Dict[str, List[Dict[str, Any]]]] = None
    expected_commands: List[PromptCommand]
    predicted_commands: Dict[str, List[PromptCommand]]
    conversation_with_diff: List[str]

    class Config:
        """Skip validation for PromptCommand protocol as pydantic does not know how to
        serialize or handle instances of a protocol.
        """

        arbitrary_types_allowed = True

    @classmethod
    def from_dialogue_understanding_test_step(
        cls,
        step: DialogueUnderstandingTestStep,
        test_case: DialogueUnderstandingTestCase,
    ) -> "FailedTestStep":
        file_path = test_case.file or ""
        user_utterance = step.text or ""
        line_number = step.line or -1

        predicted_commands: Dict[str, List[PromptCommand]] = {}
        prompts: Optional[Dict[str, List[Dict[str, Any]]]] = None
        command_generators: List[str] = []

        if step.dialogue_understanding_output:
            predicted_commands = step.dialogue_understanding_output.commands
            command_generators = step.dialogue_understanding_output.get_component_names_that_predicted_commands()  # noqa: E501
            prompts = (
                step.dialogue_understanding_output.get_component_name_to_prompt_info()
            )

        step_index = test_case.steps.index(step)

        conversation_with_diff = test_case.to_readable_conversation(
            until_step=step_index + 1
        ) + get_command_comparison(step)

        return cls(
            file=file_path,
            test_case_name=test_case.name,
            failed_user_utterance=user_utterance,
            error_line=line_number,
            pass_status=False,
            command_generators=command_generators,
            prompts=prompts,
            expected_commands=step.commands or [],
            predicted_commands=predicted_commands,
            conversation_with_diff=conversation_with_diff,
        )

    def to_dict(self, output_prompt: bool) -> Dict[str, Any]:
        step_info = {
            "file": self.file,
            "test_case": self.test_case_name,
            "failed_user_utterance": self.failed_user_utterance,
            "error_line": self.error_line,
            "pass_status": self.pass_status,
            "expected_commands": [
                command.to_dsl() for command in self.expected_commands
            ],
            "predicted_commands": [
                {
                    "component": component,
                    "commands": [command.to_dsl() for command in commands],
                }
                for component, commands in self.predicted_commands.items()
                if commands
            ],
        }

        if output_prompt and self.prompts:
            step_info["prompts"] = copy.deepcopy(self.prompts)
        elif self.prompts:
            prompts = copy.deepcopy(self.prompts)
            # remove user and system prompts
            for prompt_data in prompts.values():
                for prompt_info in prompt_data:
                    prompt_info.pop(KEY_USER_PROMPT, None)
                    prompt_info.pop(KEY_SYSTEM_PROMPT, None)

                step_info["prompts"] = prompts

        return step_info


class DialogueUnderstandingTestSuiteResult:
    """Result of a dialogue understanding test suite.

    Aggregates test results and provides metrics for the entire test suite
    used to log the results to the console and write them to a file.
    """

    def __init__(self) -> None:
        self.accuracy = {
            KEY_TEST_CASES_ACCURACY: 0.0,
            KEY_USER_UTTERANCES_ACCURACY: 0.0,
        }
        self.number_of_passed_tests = 0
        self.number_of_failed_tests = 0
        self.number_of_passed_user_utterances = 0
        self.number_of_failed_user_utterances = 0
        self.command_metrics: Optional[Dict[str, "CommandMetrics"]] = None
        self.names_of_failed_tests: List[str] = []
        self.names_of_passed_tests: List[str] = []
        self.failed_test_steps: List[FailedTestStep] = []
        self.llm_config: Optional[Dict[str, Any]] = None
        self.latency_metrics: Dict[str, float] = {}
        self.prompt_token_metrics: Dict[str, float] = {}
        self.completion_token_metrics: Dict[str, float] = {}

    @classmethod
    def from_results(
        cls,
        failing_test_results: List[DialogueUnderstandingTestResult],
        passing_test_results: List[DialogueUnderstandingTestResult],
        command_metrics: Dict[str, "CommandMetrics"],
        llm_config: Optional[Dict[str, Any]],
    ) -> "DialogueUnderstandingTestSuiteResult":
        """Create a DialogueUnderstandingTestSuiteResult object from the test results.

        Create a new instance of DialogueUnderstandingTestSuiteResult by aggregating
        metrics from passing and failing test results, along with command metrics.

        Args:
            failing_test_results: A list of DialogueUnderstandingTestResult objects
                representing the test cases that did not pass.
            passing_test_results: A list of DialogueUnderstandingTestResult objects
                representing the test cases that passed.
            command_metrics: A dictionary of command-specific performance metrics, keyed
                by command name.
            llm_config: A dictionary containing the command generator configuration.

        Returns:
            A DialogueUnderstandingTestSuiteResult object containing aggregated test
            suite metrics, including accuracy, counts of passed and failed test cases,
            user utterance statistics, and command metrics.
        """
        instance = cls()

        instance.number_of_passed_tests = len(passing_test_results)
        instance.number_of_failed_tests = len(failing_test_results)
        instance.accuracy[KEY_TEST_CASES_ACCURACY] = instance.number_of_passed_tests / (
            instance.number_of_passed_tests + instance.number_of_failed_tests
        )

        instance._set_user_utterance_metrics(failing_test_results, passing_test_results)

        instance.command_metrics = command_metrics

        instance.names_of_passed_tests = [
            passing_test_result.test_case.full_name()
            for passing_test_result in passing_test_results
        ]
        instance.names_of_failed_tests = [
            failing_test_result.test_case.full_name()
            for failing_test_result in failing_test_results
        ]

        instance.failed_test_steps = cls._create_failed_steps_from_results(
            failing_test_results
        )

        instance.latency_metrics = cls.get_latency_metrics(
            failing_test_results, passing_test_results
        )
        instance.prompt_token_metrics = cls.get_prompt_token_metrics(
            failing_test_results, passing_test_results
        )
        instance.completion_token_metrics = cls.get_completion_token_metrics(
            failing_test_results, passing_test_results
        )

        instance.llm_config = llm_config

        return instance

    def _set_user_utterance_metrics(
        self,
        failing_test_results: List[DialogueUnderstandingTestResult],
        passing_test_results: List[DialogueUnderstandingTestResult],
    ) -> None:
        # Create list of booleans indicating whether each user utterance
        # passed or failed
        user_utterances_status = [
            step.has_passed()
            for test in failing_test_results + passing_test_results
            for step in test.test_case.iterate_over_user_steps()
        ]
        # Calculate number of passed and failed user utterances
        self.number_of_passed_user_utterances = sum(user_utterances_status)
        self.number_of_failed_user_utterances = (
            len(user_utterances_status) - self.number_of_passed_user_utterances
        )
        # Calculate user utterance accuracy
        self.accuracy[KEY_USER_UTTERANCES_ACCURACY] = (
            self.number_of_passed_user_utterances
            / (
                self.number_of_failed_user_utterances
                + self.number_of_passed_user_utterances
            )
        )

    @staticmethod
    def _create_failed_steps_from_results(
        failing_test_results: List["DialogueUnderstandingTestResult"],
    ) -> List[FailedTestStep]:
        """Create list of FailedTestStep objects from failing test results.

        Given a list of failing DialogueUnderstandingTestResult objects,
        create and return a list of FailedTestStep objects for each failing user step.

        Args:
            failing_test_results: Results of failing Dialogue Understanding tests.

        Returns:
            List of aggregated FailedTestStep objects for logging to console and file.
        """
        failed_test_steps: List[FailedTestStep] = []

        for result in failing_test_results:
            test_case = result.test_case
            for step in test_case.failed_user_steps():
                failed_test_steps.append(
                    FailedTestStep.from_dialogue_understanding_test_step(
                        step, test_case
                    )
                )

        return failed_test_steps

    @staticmethod
    def _calculate_percentiles(values: List[float]) -> Dict[str, float]:
        return {
            "p50": float(np.percentile(values, 50)) if values else 0.0,
            "p90": float(np.percentile(values, 90)) if values else 0.0,
            "p99": float(np.percentile(values, 99)) if values else 0.0,
        }

    @classmethod
    def get_latency_metrics(
        cls,
        failing_test_results: List["DialogueUnderstandingTestResult"],
        passing_test_results: List["DialogueUnderstandingTestResult"],
    ) -> Dict[str, float]:
        latencies = [
            latency
            for result in failing_test_results + passing_test_results
            for step in result.test_case.steps
            for latency in step.get_latencies()
        ]

        return cls._calculate_percentiles(latencies)

    @classmethod
    def get_prompt_token_metrics(
        cls,
        failing_test_results: List["DialogueUnderstandingTestResult"],
        passing_test_results: List["DialogueUnderstandingTestResult"],
    ) -> Dict[str, float]:
        tokens = [
            token_count
            for result in failing_test_results + passing_test_results
            for step in result.test_case.steps
            for token_count in step.get_prompt_tokens()
        ]

        return cls._calculate_percentiles(tokens)

    @classmethod
    def get_completion_token_metrics(
        cls,
        failing_test_results: List["DialogueUnderstandingTestResult"],
        passing_test_results: List["DialogueUnderstandingTestResult"],
    ) -> Dict[str, float]:
        tokens = [
            token_count
            for result in failing_test_results + passing_test_results
            for step in result.test_case.steps
            for token_count in step.get_completion_tokens()
        ]

        return cls._calculate_percentiles(tokens)

    def to_dict(self, output_prompt: bool = False) -> Dict[str, Any]:
        """Builds a dictionary for writing test results to a YML file.

        Args:
            output_prompt: Whether to log the prompt or not.
        """
        # 1. Accuracy block
        result_dict: Dict[Text, Any] = {
            "accuracy": {
                "test_cases": self.accuracy[KEY_TEST_CASES_ACCURACY],
                "user_utterances": self.accuracy[KEY_USER_UTTERANCES_ACCURACY],
            },
            OUTPUT_NUMBER_OF_PASSED_TESTS: self.number_of_passed_tests,
            OUTPUT_NUMBER_OF_FAILED_TESTS: self.number_of_failed_tests,
            OUTPUT_NUMBER_OF_PASSED_USER_UTTERANCES: self.number_of_passed_user_utterances,  # noqa: E501
            OUTPUT_NUMBER_OF_FAILED_USER_UTTERANCES: self.number_of_failed_user_utterances,  # noqa: E501
        }

        cmd_metrics_output = {}
        if self.command_metrics:
            if isinstance(self.command_metrics, dict):
                for cmd_name, metrics_obj in self.command_metrics.items():
                    cmd_metrics_output[cmd_name] = metrics_obj.as_dict()
            else:
                pass

        result_dict[OUTPUT_COMMAND_METRICS] = cmd_metrics_output

        result_dict[OUTPUT_LATENCY_METRICS] = self.latency_metrics
        result_dict[OUTPUT_PROMPT_TOKEN_METRICS] = self.prompt_token_metrics
        result_dict[OUTPUT_COMPLETION_TOKEN_METRICS] = self.completion_token_metrics

        result_dict[OUTPUT_NAMES_OF_PASSED_TESTS] = self.names_of_passed_tests
        result_dict[OUTPUT_NAMES_OF_FAILED_TESTS] = self.names_of_failed_tests

        failed_steps_list = []
        for failed_test_step in self.failed_test_steps:
            failed_steps_list.append(
                failed_test_step.to_dict(output_prompt=output_prompt)
            )

        result_dict["failed_test_steps"] = failed_steps_list

        if self.llm_config:
            result_dict[OUTPUT_LLM_COMMAND_GENERATOR_CONFIG] = self.llm_config

        return result_dict



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

import structlog

from rasa.dialogue_understanding.commands.command import Command
from rasa.dialogue_understanding.commands.command_syntax_manager import (
    CommandSyntaxManager,
    CommandSyntaxVersion,
)
from rasa.dialogue_understanding.patterns.clarify import FLOW_PATTERN_CLARIFICATION
from rasa.dialogue_understanding.patterns.continue_interrupted import (
    ContinueInterruptedPatternFlowStackFrame,
)
from rasa.dialogue_understanding.stack.dialogue_stack import DialogueStack
from rasa.dialogue_understanding.stack.frames.flow_stack_frame import (
    FlowStackFrameType,
    UserFlowStackFrame,
)
from rasa.dialogue_understanding.stack.utils import (
    top_user_flow_frame,
    user_flows_on_the_stack,
)
from rasa.shared.core.events import Event, FlowInterrupted
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.trackers import DialogueStateTracker

structlogger = structlog.get_logger()


@dataclass
class StartFlowCommand(Command):
    """A command to start a flow."""

    flow: str

    @classmethod
    def command(cls) -> str:
        """Returns the command type."""
        return "start flow"

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> StartFlowCommand:
        """Converts the dictionary to a command.

        Returns:
            The converted dictionary.
        """
        try:
            return StartFlowCommand(flow=data["flow"])
        except KeyError as e:
            raise ValueError(
                f"Missing parameter '{e}' while parsing StartFlowCommand."
            ) from e

    def run_command_on_tracker(
        self,
        tracker: DialogueStateTracker,
        all_flows: FlowsList,
        original_tracker: DialogueStateTracker,
    ) -> List[Event]:
        """Runs the command on the tracker.

        Args:
            tracker: The tracker to run the command on.
            all_flows: All flows in the assistant.
            original_tracker: The tracker before any command was executed.

        Returns:
            The events to apply to the tracker.
        """
        stack = tracker.stack
        original_stack = original_tracker.stack
        applied_events: List[Event] = []

        if self.flow in user_flows_on_the_stack(stack):
            top_frame = stack.top()
            if top_frame is not None and top_frame.type() == FLOW_PATTERN_CLARIFICATION:
                return self.change_flow_frame_position_in_the_stack(stack, tracker)

            structlogger.debug(
                "command_executor.skip_command.already_started_flow", command=self
            )
            return []
        elif self.flow not in all_flows.flow_ids:
            structlogger.debug(
                "command_executor.skip_command.start_invalid_flow_id", command=self
            )
            return []

        original_user_frame = top_user_flow_frame(original_stack)
        original_top_flow = (
            original_user_frame.flow(all_flows) if original_user_frame else None
        )

        frame_type = FlowStackFrameType.REGULAR

        if original_top_flow:
            frame_type = FlowStackFrameType.INTERRUPT

            if original_user_frame is not None:
                applied_events.append(
                    FlowInterrupted(
                        original_user_frame.flow_id, original_user_frame.step_id
                    )
                )

        structlogger.debug("command_executor.start_flow", command=self)
        stack.push(UserFlowStackFrame(flow_id=self.flow, frame_type=frame_type))
        return applied_events + tracker.create_stack_updated_events(stack)

    def __hash__(self) -> int:
        return hash(self.flow)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, StartFlowCommand):
            return False

        return other.flow == self.flow

    def to_dsl(self) -> str:
        """Converts the command to a DSL string."""
        mapper = {
            CommandSyntaxVersion.v1: f"StartFlow({self.flow})",
            CommandSyntaxVersion.v2: f"start flow {self.flow}",
        }
        return mapper.get(
            CommandSyntaxManager.get_syntax_version(),
            mapper[CommandSyntaxManager.get_default_syntax_version()],
        )

    @classmethod
    def from_dsl(cls, match: re.Match, **kwargs: Any) -> Optional[StartFlowCommand]:
        """Converts the DSL string to a command."""
        return StartFlowCommand(flow=str(match.group(1).strip()))

    @staticmethod
    def regex_pattern() -> str:
        mapper = {
            CommandSyntaxVersion.v1: r"StartFlow\(['\"]?([a-zA-Z0-9_-]+)['\"]?\)",
            CommandSyntaxVersion.v2: r"^[^\w]*start flow ['\"]?([a-zA-Z0-9_-]+)['\"]?",
        }
        return mapper.get(
            CommandSyntaxManager.get_syntax_version(),
            mapper[CommandSyntaxManager.get_default_syntax_version()],
        )

    def change_flow_frame_position_in_the_stack(
        self, stack: DialogueStack, tracker: DialogueStateTracker
    ) -> List[Event]:
        """Changes the position of the flow frame in the stack.

        This is a special case when pattern clarification is the active flow and
        the same flow is selected to start. In this case, the existing flow frame
        should be moved up in the stack.
        """
        frames = stack.frames[:]

        for idx, frame in enumerate(frames):
            if isinstance(frame, UserFlowStackFrame) and frame.flow_id == self.flow:
                structlogger.debug(
                    "command_executor.change_flow_position_during_clarification",
                    command=self,
                    index=idx,
                )
                # pop the continue interrupted flow frame if it exists
                next_frame = frames[idx + 1] if idx + 1 < len(frames) else None
                if (
                    isinstance(next_frame, ContinueInterruptedPatternFlowStackFrame)
                    and next_frame.previous_flow_name == self.flow
                ):
                    stack.frames.pop(idx + 1)
                # move up the existing flow from the stack
                stack.frames.pop(idx)
                stack.push(frame)
                return tracker.create_stack_updated_events(stack)

        return []



from rasa.core.policies.enterprise_search_policy import EnterpriseSearchPolicy
from rasa.core.policies.flow_policy import FlowPolicy
from rasa.core.policies.intentless_policy import IntentlessPolicy
from rasa.core.policies.memoization import AugmentedMemoizationPolicy, MemoizationPolicy
from rasa.core.policies.rule_policy import RulePolicy
from rasa.core.policies.ted_policy import TEDPolicy
from rasa.core.policies.unexpected_intent_policy import UnexpecTEDIntentPolicy
from rasa.dialogue_understanding.coexistence.intent_based_router import (
    IntentBasedRouter,
)
from rasa.dialogue_understanding.coexistence.llm_based_router import LLMBasedRouter
from rasa.dialogue_understanding.generator import (
    LLMCommandGenerator,
)
from rasa.dialogue_understanding.generator.nlu_command_adapter import NLUCommandAdapter
from rasa.nlu.classifiers.diet_classifier import DIETClassifier
from rasa.nlu.classifiers.fallback_classifier import FallbackClassifier
from rasa.nlu.classifiers.keyword_intent_classifier import KeywordIntentClassifier
from rasa.nlu.classifiers.logistic_regression_classifier import (
    LogisticRegressionClassifier,
)
from rasa.nlu.classifiers.mitie_intent_classifier import MitieIntentClassifier
from rasa.nlu.classifiers.sklearn_intent_classifier import SklearnIntentClassifier
from rasa.nlu.extractors.crf_entity_extractor import CRFEntityExtractor
from rasa.nlu.extractors.duckling_entity_extractor import DucklingEntityExtractor
from rasa.nlu.extractors.entity_synonyms import EntitySynonymMapper
from rasa.nlu.extractors.mitie_entity_extractor import MitieEntityExtractor
from rasa.nlu.extractors.regex_entity_extractor import RegexEntityExtractor
from rasa.nlu.extractors.spacy_entity_extractor import SpacyEntityExtractor
from rasa.nlu.featurizers.dense_featurizer.convert_featurizer import ConveRTFeaturizer
from rasa.nlu.featurizers.dense_featurizer.lm_featurizer import LanguageModelFeaturizer
from rasa.nlu.featurizers.dense_featurizer.mitie_featurizer import MitieFeaturizer
from rasa.nlu.featurizers.dense_featurizer.spacy_featurizer import SpacyFeaturizer
from rasa.nlu.featurizers.sparse_featurizer.count_vectors_featurizer import (
    CountVectorsFeaturizer,
)
from rasa.nlu.featurizers.sparse_featurizer.lexical_syntactic_featurizer import (
    LexicalSyntacticFeaturizer,
)
from rasa.nlu.featurizers.sparse_featurizer.regex_featurizer import RegexFeaturizer
from rasa.nlu.selectors.response_selector import ResponseSelector
from rasa.nlu.tokenizers.jieba_tokenizer import JiebaTokenizer
from rasa.nlu.tokenizers.mitie_tokenizer import MitieTokenizer
from rasa.nlu.tokenizers.spacy_tokenizer import SpacyTokenizer
from rasa.nlu.tokenizers.whitespace_tokenizer import WhitespaceTokenizer
from rasa.nlu.utils.mitie_utils import MitieNLP
from rasa.nlu.utils.spacy_utils import SpacyNLP

DEFAULT_COMPONENTS = [
    # Message Classifiers
    DIETClassifier,
    FallbackClassifier,
    KeywordIntentClassifier,
    MitieIntentClassifier,
    SklearnIntentClassifier,
    LogisticRegressionClassifier,
    NLUCommandAdapter,
    LLMCommandGenerator,
    LLMBasedRouter,
    IntentBasedRouter,
    # Response Selectors
    ResponseSelector,
    # Message Entity Extractors
    CRFEntityExtractor,
    DucklingEntityExtractor,
    EntitySynonymMapper,
    MitieEntityExtractor,
    SpacyEntityExtractor,
    RegexEntityExtractor,
    # Message Feauturizers
    LexicalSyntacticFeaturizer,
    ConveRTFeaturizer,
    MitieFeaturizer,
    SpacyFeaturizer,
    CountVectorsFeaturizer,
    LanguageModelFeaturizer,
    RegexFeaturizer,
    # Tokenizers
    JiebaTokenizer,
    MitieTokenizer,
    SpacyTokenizer,
    WhitespaceTokenizer,
    # Language Model Providers
    MitieNLP,
    SpacyNLP,
    # Dialogue Management Policies
    TEDPolicy,
    UnexpecTEDIntentPolicy,
    RulePolicy,
    MemoizationPolicy,
    AugmentedMemoizationPolicy,
    FlowPolicy,
    EnterpriseSearchPolicy,
    IntentlessPolicy,
]



import asyncio
from typing import Any, Dict, List, Optional, Text

import structlog
from tqdm import tqdm

from rasa.core.channels import CollectingOutputChannel, UserMessage
from rasa.core.exceptions import AgentNotReady
from rasa.core.persistor import StorageType
from rasa.core.utils import AvailableEndpoints
from rasa.dialogue_understanding.commands import Command
from rasa.dialogue_understanding.utils import set_record_commands_and_prompts
from rasa.dialogue_understanding_test.du_test_case import (
    DialogueUnderstandingOutput,
    DialogueUnderstandingTestCase,
    DialogueUnderstandingTestStep,
)
from rasa.dialogue_understanding_test.du_test_result import (
    DialogueUnderstandingTestResult,
)
from rasa.dialogue_understanding_test.test_case_simulation.test_case_tracker_simulator import (  # noqa: E501
    TestCaseTrackerSimulator,
    TestCaseTrackerSimulatorException,
    TestCaseTrackerSimulatorResult,
)
from rasa.dialogue_understanding_test.utils import filter_metadata
from rasa.e2e_test.e2e_test_case import (
    KEY_STUB_CUSTOM_ACTIONS,
    Fixture,
    Metadata,
)
from rasa.e2e_test.e2e_test_runner import E2ETestRunner
from rasa.shared.core.events import UserUttered
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.nlu.constants import PREDICTED_COMMANDS, PROMPTS
from rasa.utils.endpoints import EndpointConfig

structlogger = structlog.get_logger()


class DialogueUnderstandingTestRunner:
    """Dialogue Understanding test suite runner.

    Responsible for executing dialogue understanding test cases by simulating
    conversations and validating expected outputs against actual responses.

    Attributes:
        agent: The Rasa agent instance used for processing messages
    """

    def __init__(
        self,
        model_path: Optional[Text] = None,
        model_server: Optional[EndpointConfig] = None,
        remote_storage: Optional[StorageType] = None,
        endpoints: Optional[AvailableEndpoints] = None,
    ) -> None:
        """Initializes the Dialogue Understanding test suite runner.

        Args:
            model_path: Path to the model.
            model_server: Model server configuration.
            remote_storage: Remote storage to use for model retrieval.
            endpoints: Endpoints configuration.
        """
        import rasa.core.agent

        self._check_action_server(endpoints)

        self.agent = asyncio.run(
            rasa.core.agent.load_agent(
                model_path=model_path,
                model_server=model_server,
                remote_storage=remote_storage,
                endpoints=endpoints,
            )
        )
        if not self.agent.is_ready():
            raise AgentNotReady(
                "Agent needs to be prepared before usage. "
                "Please check that the agent was able to "
                "load the trained model."
            )

    @staticmethod
    def _check_action_server(endpoints: AvailableEndpoints) -> None:
        """Check if the action server is reachable."""
        are_custom_actions_stubbed = (
            endpoints
            and endpoints.action
            and endpoints.action.kwargs.get(KEY_STUB_CUSTOM_ACTIONS)
        )
        if endpoints and not are_custom_actions_stubbed:
            E2ETestRunner._action_server_is_reachable(
                endpoints, "dialogue_understanding_test_runner"
            )

    async def run_test_cases(
        self,
        test_cases: List[DialogueUnderstandingTestCase],
        fixtures: List[Fixture],
        metadata: List[Metadata],
    ) -> List[DialogueUnderstandingTestResult]:
        """Run the dialogue understanding tests.

        Args:
            test_cases: List of test cases.
            fixtures: List of fixtures.
            metadata: List of metadata.

        Returns:
            List[DialogueUnderstandingTestResult]: List of test results.
        """
        results = []
        output_channel = CollectingOutputChannel()

        structlogger.info("Starting dialogue understanding tests.")

        for i in tqdm(range(len(test_cases))):
            test_case = test_cases[i]

            # set up the tracker by simulating the conversation
            tracker_simulator = TestCaseTrackerSimulator(self.agent, test_case)
            await tracker_simulator.initialize_tracker(fixtures)
            try:
                simulation_result = await tracker_simulator.simulate_test_case(metadata)
            except TestCaseTrackerSimulatorException as e:
                structlogger.error(
                    "Failed to simulate test case. Skipping test case.",
                    error=str(e),
                )
                continue

            # run the actual test case
            test_result = await self.run_test_case(
                test_case, metadata, simulation_result, output_channel
            )
            results.append(test_result)

        structlogger.info("Finished dialogue understanding tests.")

        return results

    async def run_test_case(
        self,
        test_case: DialogueUnderstandingTestCase,
        metadata: List[Metadata],
        simulation_result: TestCaseTrackerSimulatorResult,
        output_channel: CollectingOutputChannel,
    ) -> DialogueUnderstandingTestResult:
        """Runs dialogue understanding test case.

        Args:
            test_case: The test case to run.
            metadata: List of metadata.
            simulation_result: The test case tracker.
            output_channel: The output channel.

        Returns:
            A dialogue understanding test result.
        """
        test_passed = True

        sender_id = simulation_result.sender_id
        user_uttered_event_indices = simulation_result.user_uttered_event_indices
        test_case_tracker = await self.agent.tracker_store.retrieve(sender_id)

        for user_step_index, user_step in enumerate(
            test_case.iterate_over_user_steps()
        ):
            if user_step_index >= len(user_uttered_event_indices):
                # the conversation could not be simulated completely and was
                # stopped early, no further user messages can be evaluated
                return DialogueUnderstandingTestResult(
                    test_case=test_case, passed=test_passed
                )

            # create and save the tracker at the time just
            # before the user message was sent
            step_sender_id = f"{sender_id}_{user_step_index}"
            await self._create_tracker_for_user_step(
                step_sender_id,
                test_case_tracker,
                user_uttered_event_indices[user_step_index],
            )

            # send the user message
            try:
                await self._send_user_message(
                    step_sender_id,
                    test_case,
                    user_step,
                    metadata,
                    output_channel=output_channel,
                )
            except Exception as e:
                structlogger.error(
                    "dialogue_understanding_test_runner.send_user_message.failed",
                    test_case=test_case.full_name(),
                    user_message=user_step.text,
                    error=str(e),
                )
                # as sending the user message failed, we cannot continue with the test
                # as subsequent user message do not match the actual conversation
                # return a test result up until this point
                return DialogueUnderstandingTestResult(
                    test_case=test_case, passed=test_passed
                )

            # get the dialogue understanding output
            tracker = await self.agent.tracker_store.retrieve(step_sender_id)
            dialogue_understanding_output = self.get_dialogue_understanding_output(
                tracker, user_uttered_event_indices[user_step_index]
            )
            user_step.dialogue_understanding_output = dialogue_understanding_output

            # check if we have a command match
            if not user_step.has_passed():
                test_passed = False

        return DialogueUnderstandingTestResult(test_case=test_case, passed=test_passed)

    def get_dialogue_understanding_output(
        self,
        tracker: DialogueStateTracker,
        index_user_uttered_event: int,
    ) -> Optional[DialogueUnderstandingOutput]:
        """Returns the dialogue understanding output.

        Creates the dialogue understanding output from the commands and prompts
        added to the user uttered event.

        Args:
            tracker: The tracker for the current test step.
            index_user_uttered_event: The index of the user uttered event.

        Returns:
        The dialogue understanding output with commands and optionally prompts.
        """
        user_uttered_event = self._get_user_uttered_event_from_tracker(
            tracker, index_user_uttered_event
        )
        if user_uttered_event is None:
            return None

        predicted_commands: Dict[str, Any] = user_uttered_event.parse_data.get(  # type:ignore[assignment]
            PREDICTED_COMMANDS, {}
        )
        # convert the predicted commands to Command objects
        commands = {}
        for component, list_of_commands in predicted_commands.items():
            # remove any duplicate commands
            commands[component] = list(
                set(
                    [Command.command_from_json(command) for command in list_of_commands]
                )
            )

        return DialogueUnderstandingOutput(
            commands=commands,
            prompts=user_uttered_event.parse_data.get(PROMPTS, []),
        )

    @staticmethod
    def _get_user_uttered_event_from_tracker(
        tracker: DialogueStateTracker,
        index_user_uttered_event: int,
    ) -> Optional[UserUttered]:
        """Returns the user uttered event from the tracker at the given index."""
        if (
            tracker.events is None
            or len(list(tracker.events)) <= index_user_uttered_event
        ):
            return None

        user_uttered_event = list(tracker.events)[index_user_uttered_event]

        if not user_uttered_event or not isinstance(user_uttered_event, UserUttered):
            return None

        return user_uttered_event

    async def _create_tracker_for_user_step(
        self,
        step_sender_id: str,
        test_case_tracker: DialogueStateTracker,
        index_user_uttered_event: int,
    ) -> None:
        """Creates a tracker for the user step."""
        tracker = test_case_tracker.copy()
        # modify the sender id so that the test case tracker is not overwritten
        tracker.sender_id = step_sender_id

        if tracker.events:
            # get timestamp of the event just before the user uttered event
            timestamp = tracker.events[index_user_uttered_event - 1].timestamp
            # revert the tracker to the event just before the user uttered event
            tracker = tracker.travel_back_in_time(timestamp)

        # store the tracker with the unique sender id
        await self.agent.tracker_store.save(tracker)

    async def _send_user_message(
        self,
        sender_id: str,
        test_case: DialogueUnderstandingTestCase,
        user_step: DialogueUnderstandingTestStep,
        metadata: List[Metadata],
        output_channel: CollectingOutputChannel = CollectingOutputChannel(),
    ) -> None:
        """Sends a user message to the agent."""
        metadata_for_step = filter_metadata(test_case, user_step, metadata, sender_id)

        user_message = UserMessage(
            user_step.text,
            output_channel,
            sender_id,
            metadata=metadata_for_step,
        )

        with set_record_commands_and_prompts():
            await self.agent.handle_message(user_message)



from contextlib import contextmanager
from typing import Any, Dict, Generator, List, Optional, Text

from rasa.dialogue_understanding.commands import Command
from rasa.dialogue_understanding.constants import (
    RASA_RECORD_COMMANDS_AND_PROMPTS_ENV_VAR_NAME,
)
from rasa.shared.constants import ROUTE_TO_CALM_SLOT
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.nlu.constants import (
    COMMANDS,
    KEY_COMPONENT_NAME,
    KEY_LLM_RESPONSE_METADATA,
    KEY_PROMPT_NAME,
    KEY_SYSTEM_PROMPT,
    KEY_USER_PROMPT,
    PREDICTED_COMMANDS,
    PROMPTS,
    SET_SLOT_COMMAND,
)
from rasa.shared.nlu.training_data.message import Message
from rasa.shared.providers.llm.llm_response import LLMResponse
from rasa.utils.common import get_bool_env_variable

record_commands_and_prompts = get_bool_env_variable(
    RASA_RECORD_COMMANDS_AND_PROMPTS_ENV_VAR_NAME, False
)


@contextmanager
def set_record_commands_and_prompts() -> Generator:
    global record_commands_and_prompts
    record_commands_and_prompts = True
    try:
        yield
    finally:
        record_commands_and_prompts = False


def add_commands_to_message_parse_data(
    message: Message, component_name: str, commands: List[Command]
) -> None:
    """Add commands to the message parse data.

    Commands are only added in case the flag 'record_commands_and_prompts' is set.
    Example of predicted commands in the message parse data:
        Message(data={
          PREDICTED_COMMANDS: {
            "MultiStepLLMCommandGenerator": [
                {"command": "set_slot", "name": "slot_name", "value": "slot_value"},
            ],
            "NLUCommandAdapter": [
                {"command": "start_flow", "name": "test_flow"},
            ]
          }
        })
    """
    # only set commands if the flag "record_commands_and_prompts" is set to True
    if not record_commands_and_prompts:
        return

    commands_as_dict = [command.as_dict() for command in commands]

    if message.get(PREDICTED_COMMANDS) is not None:
        predicted_commands = message.get(PREDICTED_COMMANDS)
        if component_name in predicted_commands:
            predicted_commands[component_name].extend(commands_as_dict)
        else:
            predicted_commands[component_name] = commands_as_dict
    else:
        predicted_commands = {component_name: commands_as_dict}

    message.set(
        PREDICTED_COMMANDS,
        predicted_commands,
        add_to_output=True,
    )


def add_prompt_to_message_parse_data(
    message: Message,
    component_name: str,
    prompt_name: str,
    user_prompt: str,
    system_prompt: Optional[str] = None,
    llm_response: Optional[LLMResponse] = None,
) -> None:
    """Add prompt to the message parse data.

    Prompt is only added in case the flag 'record_commands_and_prompts' is set.
    Example of prompts in the message parse data:
        Message(data={
            PROMPTS: [
                {
                    "component_name": "MultiStepLLMCommandGenerator",
                    "prompt_name": "fill_slots_prompt",
                    "user_prompt": "...",
                    "system_prompt": "...",
                    "llm_response_metadata": { ... }
                },
                {
                    "component_name": "MultiStepLLMCommandGenerator",
                    "prompt_name": "handle_flows_prompt",
                    "user_prompt": "...",
                    "system_prompt": "...",
                    "llm_response_metadata": { ... }
                },
                {
                    "component_name": "SingleStepLLMCommandGenerator",
                    "prompt_name": "prompt_template",
                    "user_prompt": "...",
                    "system_prompt": "...",
                    "llm_response_metadata": { ... }
                }
            ]
        }
    )
    """
    # Only set prompt if the flag "record_commands_and_prompts" is set to True.
    if not record_commands_and_prompts:
        return

    # Construct the dictionary with prompt details.
    prompt_data: Dict[Text, Any] = {
        KEY_COMPONENT_NAME: component_name,
        KEY_PROMPT_NAME: prompt_name,
        KEY_USER_PROMPT: user_prompt,
        KEY_LLM_RESPONSE_METADATA: llm_response.to_dict() if llm_response else None,
        **({KEY_SYSTEM_PROMPT: system_prompt} if system_prompt else {}),
    }

    # Get or create a top-level "prompts" list.
    prompts = message.get(PROMPTS) or []
    prompts.append(prompt_data)

    # Update the message with the new prompts list.
    message.set(PROMPTS, prompts, add_to_output=True)


def _handle_via_nlu_in_coexistence(
    tracker: Optional[DialogueStateTracker], message: Message
) -> bool:
    """Check if the message should be handled by the NLU subsystem in coexistence mode."""  # noqa: E501
    if not tracker:
        return False

    if not tracker.has_coexistence_routing_slot:
        return False

    value = tracker.get_slot(ROUTE_TO_CALM_SLOT)
    if value is not None:
        return not value

    # routing slot has been reset so we need to check
    # the command issued by the Router component
    if message.get(COMMANDS):
        for command in message.get(COMMANDS):
            if (
                command.get("command") == SET_SLOT_COMMAND
                and command.get("name") == ROUTE_TO_CALM_SLOT
            ):
                return not command.get("value")

    return False



from __future__ import annotations

from typing import Any, Dict, List, Optional

import structlog

from rasa.core.actions.action import Action, create_bot_utterance
from rasa.core.channels import OutputChannel
from rasa.core.nlg import NaturalLanguageGenerator
from rasa.core.utils import add_bot_utterance_metadata
from rasa.dialogue_understanding.patterns.continue_interrupted import (
    ContinueInterruptedPatternFlowStackFrame,
)
from rasa.dialogue_understanding.patterns.handle_digressions import (
    HandleDigressionsPatternFlowStackFrame,
)
from rasa.dialogue_understanding.stack.frames.flow_stack_frame import (
    FlowStackFrameType,
    UserFlowStackFrame,
)
from rasa.shared.core.constants import (
    ACTION_BLOCK_DIGRESSION,
    ACTION_CONTINUE_DIGRESSION,
)
from rasa.shared.core.domain import Domain
from rasa.shared.core.events import Event, FlowInterrupted
from rasa.shared.core.trackers import DialogueStateTracker

structlogger = structlog.get_logger()


class ActionBlockDigressions(Action):
    """Action which blocks an interruption and continues the current flow."""

    def name(self) -> str:
        """Return the action name."""
        return ACTION_BLOCK_DIGRESSION

    async def run(
        self,
        output_channel: OutputChannel,
        nlg: NaturalLanguageGenerator,
        tracker: DialogueStateTracker,
        domain: Domain,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> List[Event]:
        """Update the stack."""
        structlogger.debug("action_block_digressions.run")
        top_frame = tracker.stack.top()

        if not isinstance(top_frame, HandleDigressionsPatternFlowStackFrame):
            return []

        blocked_flow_id = top_frame.interrupting_flow_id
        frame_type = FlowStackFrameType.REGULAR

        stack = tracker.stack
        stack.push(
            UserFlowStackFrame(flow_id=blocked_flow_id, frame_type=frame_type), 0
        )
        stack.push(
            ContinueInterruptedPatternFlowStackFrame(
                previous_flow_name=blocked_flow_id
            ),
            1,
        )
        events = tracker.create_stack_updated_events(stack)

        utterance = "utter_block_digressions"
        message = await nlg.generate(
            utterance,
            tracker,
            output_channel.name(),
        )

        if message is None:
            structlogger.error(
                "action_block_digressions.run.failed.finding.utter",
                utterance=utterance,
            )
        else:
            message = add_bot_utterance_metadata(
                message, utterance, nlg, domain, tracker
            )
            events.append(create_bot_utterance(message))

        return events


class ActionContinueDigression(Action):
    """Action which continues with an interruption."""

    def name(self) -> str:
        """Return the action name."""
        return ACTION_CONTINUE_DIGRESSION

    async def run(
        self,
        output_channel: OutputChannel,
        nlg: NaturalLanguageGenerator,
        tracker: DialogueStateTracker,
        domain: Domain,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> List[Event]:
        """Update the stack."""
        structlogger.debug("action_continue_digression.run")
        top_frame = tracker.stack.top()

        if not isinstance(top_frame, HandleDigressionsPatternFlowStackFrame):
            return []

        blocked_flow_id = top_frame.interrupting_flow_id
        frame_type = FlowStackFrameType.INTERRUPT
        stack = tracker.stack
        stack.push(UserFlowStackFrame(flow_id=blocked_flow_id, frame_type=frame_type))

        events = [
            FlowInterrupted(
                flow_id=top_frame.interrupted_flow_id,
                step_id=top_frame.interrupted_step_id,
            )
        ] + tracker.create_stack_updated_events(stack)

        utterance = "utter_continue_interruption"
        message = await nlg.generate(
            utterance,
            tracker,
            output_channel.name(),
        )

        if message is None:
            structlogger.error(
                "action_continue_digression.run.failed.finding.utter",
                utterance=utterance,
            )
        else:
            message = add_bot_utterance_metadata(
                message, utterance, nlg, domain, tracker
            )
            events.append(create_bot_utterance(message))

        return events



from __future__ import annotations

from typing import Any, Dict, List, Text

import rasa.dialogue_understanding.processor.command_processor
from rasa.engine.graph import ExecutionContext, GraphComponent
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.shared.core.events import Event
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.core.training_data.structures import StoryGraph


class CommandProcessorComponent(GraphComponent):
    """Processes commands by issuing events to modify a tracker.

    Minimal component that applies commands to a tracker."""

    def __init__(self, execution_context: ExecutionContext):
        self._execution_context = execution_context

    @classmethod
    def create(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
    ) -> CommandProcessorComponent:
        """Creates component (see parent class for full docstring)."""
        return cls(execution_context)

    def execute_commands(
        self,
        tracker: DialogueStateTracker,
        flows: FlowsList,
        story_graph: StoryGraph,
    ) -> List[Event]:
        """Execute commands to update tracker state."""
        return rasa.dialogue_understanding.processor.command_processor.execute_commands(
            tracker, flows, self._execution_context, story_graph
        )



import argparse
import asyncio
import sys
from pathlib import Path
from typing import Dict, List, Optional, Text, Union

import structlog

import rasa.cli.arguments.train as train_arguments
import rasa.cli.utils
import rasa.core.utils
import rasa.utils.common
from rasa.cli import SubParsersAction
from rasa.core import ContextualResponseRephraser
from rasa.core.nlg.generator import NaturalLanguageGenerator
from rasa.core.train import do_compare_training
from rasa.engine.validation import validate_api_type_config_key_usage
from rasa.shared.constants import (
    CONFIG_MANDATORY_KEYS,
    CONFIG_MANDATORY_KEYS_CORE,
    CONFIG_MANDATORY_KEYS_NLU,
    DEFAULT_DATA_PATH,
    DEFAULT_DOMAIN_PATHS,
    LLM_CONFIG_KEY,
)
from rasa.shared.importers.importer import TrainingDataImporter

structlogger = structlog.getLogger(__name__)


def add_subparser(
    subparsers: SubParsersAction, parents: List[argparse.ArgumentParser]
) -> None:
    """Add all training parsers.

    Args:
        subparsers: subparser we are going to attach to
        parents: Parent parsers, needed to ensure tree structure in argparse
    """
    train_parser = subparsers.add_parser(
        "train",
        help="Trains a Rasa model using your NLU data and stories.",
        parents=parents,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    train_arguments.set_train_arguments(train_parser)

    train_subparsers = train_parser.add_subparsers()
    train_core_parser = train_subparsers.add_parser(
        "core",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Trains a Rasa Core model using your stories.",
    )
    train_core_parser.set_defaults(func=run_core_training)

    train_nlu_parser = train_subparsers.add_parser(
        "nlu",
        parents=parents,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Trains a Rasa NLU model using your NLU data.",
    )
    train_nlu_parser.set_defaults(func=run_nlu_training)

    train_parser.set_defaults(func=lambda args: run_training(args, can_exit=True))

    train_arguments.set_train_core_arguments(train_core_parser)
    train_arguments.set_train_nlu_arguments(train_nlu_parser)


def _check_nlg_endpoint_validity(endpoint: Union[Path, str]) -> None:
    try:
        endpoints = rasa.core.utils.read_endpoints_from_path(endpoint)
        if endpoints.nlg is not None:
            validate_api_type_config_key_usage(
                endpoints.nlg.kwargs,
                LLM_CONFIG_KEY,
                ContextualResponseRephraser.__name__,
            )
        NaturalLanguageGenerator.create(endpoints.nlg)
    except Exception as e:
        structlogger.error(
            "cli.train.nlg_failed_to_initialise.validation_error",
            exception=f"{e}",
            event_info=(
                f"The validation failed for NLG configuration defined in "
                f"{endpoint}. Please make sure the NLG configuration is correct."
            ),
        )
        sys.exit(1)


def run_training(args: argparse.Namespace, can_exit: bool = False) -> Optional[Text]:
    """Trains a model.

    Args:
        args: Namespace arguments.
        can_exit: If `True`, the operation can send `sys.exit` in the case
            training was not successful.

    Returns:
        Path to a trained model or `None` if training was not successful.
    """
    from rasa.api import train as train_all

    domain = rasa.cli.utils.get_validated_path(
        args.domain, "domain", DEFAULT_DOMAIN_PATHS, none_is_valid=True
    )
    config = rasa.cli.utils.get_validated_config(args.config, CONFIG_MANDATORY_KEYS)

    _check_nlg_endpoint_validity(args.endpoints)

    training_files = [
        rasa.cli.utils.get_validated_path(
            f, "data", DEFAULT_DATA_PATH, none_is_valid=True
        )
        for f in args.data
    ]

    training_data_importer = TrainingDataImporter.load_from_config(
        domain_path=domain, training_data_paths=args.data, config_path=config
    )

    if not args.skip_validation:
        structlogger.info(
            "cli.train.run_training",
            event_info="Started validating domain and training data...",
        )

        rasa.cli.utils.validate_files(
            args.fail_on_validation_warnings,
            args.validation_max_history,
            training_data_importer,
        )

    training_result = train_all(
        domain=domain,
        config=config,
        training_files=training_files,
        output=args.out,
        dry_run=args.dry_run,
        force_training=args.force,
        fixed_model_name=args.fixed_model_name,
        persist_nlu_training_data=args.persist_nlu_data,
        core_additional_arguments={
            **extract_core_additional_arguments(args),
        },
        nlu_additional_arguments=extract_nlu_additional_arguments(args),
        model_to_finetune=_model_for_finetuning(args),
        finetuning_epoch_fraction=args.epoch_fraction,
        remote_storage=args.remote_storage,
        file_importer=training_data_importer,
        keep_local_model_copy=args.keep_local_model_copy,
    )
    if training_result.code != 0 and can_exit:
        sys.exit(training_result.code)

    return training_result.model


def _model_for_finetuning(args: argparse.Namespace) -> Optional[Text]:
    if args.finetune == train_arguments.USE_LATEST_MODEL_FOR_FINE_TUNING:
        # We use this constant to signal that the user specified `--finetune` but
        # didn't provide a path to a model. In this case we try to load the latest
        # model from the output directory (that's usually models/).
        return args.out
    else:
        return args.finetune


def run_core_training(args: argparse.Namespace) -> Optional[Text]:
    """Trains a Rasa Core model only.

    Args:
        args: Command-line arguments to configure training.

    Returns:
        Path to a trained model or `None` if training was not successful.
    """
    from rasa.model_training import train_core

    args.domain = rasa.cli.utils.get_validated_path(
        args.domain, "domain", DEFAULT_DOMAIN_PATHS, none_is_valid=True
    )
    story_file = rasa.cli.utils.get_validated_path(
        args.stories, "stories", DEFAULT_DATA_PATH, none_is_valid=True
    )
    additional_arguments = {
        **extract_core_additional_arguments(args),
    }

    # Policies might be a list for the compare training. Do normal training
    # if only list item was passed.
    if not isinstance(args.config, list) or len(args.config) == 1:
        if isinstance(args.config, list):
            args.config = args.config[0]

        config = rasa.cli.utils.get_validated_config(
            args.config, CONFIG_MANDATORY_KEYS_CORE
        )

        return asyncio.run(
            train_core(
                domain=args.domain,
                config=config,
                stories=story_file,
                output=args.out,
                fixed_model_name=args.fixed_model_name,
                additional_arguments=additional_arguments,
                model_to_finetune=_model_for_finetuning(args),
                finetuning_epoch_fraction=args.epoch_fraction,
                keep_local_model_copy=args.keep_local_model_copy,
            )
        )
    else:
        asyncio.run(do_compare_training(args, story_file, additional_arguments))
        return None


def run_nlu_training(args: argparse.Namespace) -> Optional[Text]:
    """Trains an NLU model.

    Args:
        args: Namespace arguments.

    Returns:
        Path to a trained model or `None` if training was not successful.
    """
    from rasa.model_training import train_nlu

    config = rasa.cli.utils.get_validated_config(args.config, CONFIG_MANDATORY_KEYS_NLU)
    nlu_data = rasa.cli.utils.get_validated_path(
        args.nlu, "nlu", DEFAULT_DATA_PATH, none_is_valid=True
    )

    if args.domain:
        args.domain = rasa.cli.utils.get_validated_path(
            args.domain, "domain", DEFAULT_DOMAIN_PATHS, none_is_valid=True
        )

    return asyncio.run(
        train_nlu(
            config=config,
            nlu_data=nlu_data,
            output=args.out,
            fixed_model_name=args.fixed_model_name,
            persist_nlu_training_data=args.persist_nlu_data,
            additional_arguments={
                **extract_nlu_additional_arguments(args),
            },
            domain=args.domain,
            model_to_finetune=_model_for_finetuning(args),
            finetuning_epoch_fraction=args.epoch_fraction,
            keep_local_model_copy=args.keep_local_model_copy,
        )
    )


def extract_core_additional_arguments(args: argparse.Namespace) -> Dict:
    arguments = {}

    if "augmentation" in args:
        arguments["augmentation_factor"] = args.augmentation
    if "debug_plots" in args:
        arguments["debug_plots"] = args.debug_plots

    return arguments


def extract_nlu_additional_arguments(args: argparse.Namespace) -> Dict:
    arguments = {}

    if "num_threads" in args:
        arguments["num_threads"] = args.num_threads

    return arguments



import base64
import json
import os
from dataclasses import dataclass
from typing import AsyncIterator, Dict, Optional

import aiohttp
import structlog
from aiohttp import ClientConnectorError, ClientTimeout

from rasa.core.channels.voice_stream.audio_bytes import HERTZ, RasaAudioBytes
from rasa.core.channels.voice_stream.tts.tts_engine import (
    TTSEngine,
    TTSEngineConfig,
    TTSError,
)
from rasa.shared.constants import CARTESIA_API_KEY_ENV_VAR
from rasa.shared.exceptions import ConnectionException

structlogger = structlog.get_logger()


@dataclass
class CartesiaTTSConfig(TTSEngineConfig):
    model_id: Optional[str] = None
    version: Optional[str] = None
    endpoint: Optional[str] = None


class CartesiaTTS(TTSEngine[CartesiaTTSConfig]):
    session: Optional[aiohttp.ClientSession] = None
    required_env_vars = (CARTESIA_API_KEY_ENV_VAR,)

    def __init__(self, config: Optional[CartesiaTTSConfig] = None):
        super().__init__(config)
        timeout = ClientTimeout(total=self.config.timeout)
        # Have to create this class-shared session lazily at run time otherwise
        # the async event loop doesn't work
        if self.__class__.session is None or self.__class__.session.closed:
            self.__class__.session = aiohttp.ClientSession(timeout=timeout)

    @staticmethod
    def get_request_body(text: str, config: CartesiaTTSConfig) -> Dict:
        """Create the request body for cartesia."""
        # more info on payload:
        # https://docs.cartesia.ai/reference/api-reference/rest/stream-speech-bytes
        return {
            "model_id": config.model_id,
            "transcript": text,
            "language": config.language,
            "voice": {
                "mode": "id",
                "id": config.voice,
            },
            "output_format": {
                "container": "raw",
                "encoding": "pcm_mulaw",
                "sample_rate": HERTZ,
            },
        }

    @staticmethod
    def get_request_headers(config: CartesiaTTSConfig) -> dict[str, str]:
        cartesia_api_key = os.environ[CARTESIA_API_KEY_ENV_VAR]
        return {
            "Cartesia-Version": str(config.version),
            "Content-Type": "application/json",
            "X-API-Key": str(cartesia_api_key),
        }

    async def synthesize(
        self, text: str, config: Optional[CartesiaTTSConfig] = None
    ) -> AsyncIterator[RasaAudioBytes]:
        """Generate speech from text using a remote TTS system."""
        config = self.config.merge(config)
        payload = self.get_request_body(text, config)
        headers = self.get_request_headers(config)
        url = self.config.endpoint
        if self.session is None:
            raise ConnectionException("Client session is not initialized")
        try:
            async with self.session.post(
                url, headers=headers, json=payload, chunked=True
            ) as response:
                if 200 <= response.status < 300:
                    async for chunk in response.content:
                        # we are looking for chunks in the response that look like
                        # b"data: {..., data: <base64 encoded audio bytes> ...}"
                        # and extract the audio bytes from that
                        if chunk.startswith(b"data: "):
                            json_bytes = chunk[5:-1]
                            json_data = json.loads(json_bytes.decode())
                            if "data" in json_data:
                                base64_encoded_bytes = json_data["data"]
                                channel_bytes = base64.b64decode(base64_encoded_bytes)
                                yield self.engine_bytes_to_rasa_audio_bytes(
                                    channel_bytes
                                )
                    return
                elif response.status == 401:
                    structlogger.error(
                        "cartesia.synthesize.rest.unauthorized",
                        status_code=response.status,
                    )
                    raise TTSError(
                        "Unauthorized. Please make sure you have the correct API key."
                    )
                else:
                    response_text = await response.text()
                    structlogger.error(
                        "cartesia.synthesize.rest.failed",
                        status_code=response.status,
                        msg=response_text,
                    )
                    raise TTSError(f"TTS failed: {response_text}")
        except ClientConnectorError as e:
            raise TTSError(e)
        except TimeoutError as e:
            raise TTSError(e)

    def engine_bytes_to_rasa_audio_bytes(self, chunk: bytes) -> RasaAudioBytes:
        """Convert the generated tts audio bytes into rasa audio bytes."""
        return RasaAudioBytes(chunk)

    @staticmethod
    def get_default_config() -> CartesiaTTSConfig:
        return CartesiaTTSConfig(
            language="en",
            voice="248be419-c632-4f23-adf1-5324ed7dbf1d",
            timeout=10,
            model_id="sonic-english",
            version="2024-06-10",
            endpoint="https://api.cartesia.ai/tts/sse",
        )

    @classmethod
    def from_config_dict(cls, config: Dict) -> "CartesiaTTS":
        return cls(CartesiaTTSConfig.from_dict(config))



import sys
import time
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional, Text, Union

import randomname
import structlog

import rasa.engine.validation
import rasa.model
import rasa.shared.constants
import rasa.shared.exceptions
import rasa.shared.utils.cli
import rasa.shared.utils.common
import rasa.shared.utils.io
import rasa.utils.common
from rasa import telemetry
from rasa.core.persistor import StorageType
from rasa.engine.caching import LocalTrainingCache
from rasa.engine.recipes.recipe import Recipe
from rasa.engine.runner.dask import DaskGraphRunner
from rasa.engine.storage.local_model_storage import LocalModelStorage
from rasa.engine.storage.storage import ModelStorage
from rasa.engine.training.components import FingerprintStatus
from rasa.engine.training.graph_trainer import GraphTrainer
from rasa.shared.constants import CONFIG_RECIPE_KEY
from rasa.shared.core.domain import Domain
from rasa.shared.core.events import SlotSet
from rasa.shared.core.training_data.structures import StoryGraph
from rasa.shared.data import TrainingType
from rasa.shared.exceptions import RasaException
from rasa.shared.importers.importer import TrainingDataImporter

CODE_NEEDS_TO_BE_RETRAINED = 0b0001
CODE_FORCED_TRAINING = 0b1000
CODE_NO_NEED_TO_TRAIN = 0b0000

structlogger = structlog.get_logger()


class TrainingResult(NamedTuple):
    """Holds information about the results of training."""

    model: Optional[Text] = None
    code: int = 0
    dry_run_results: Optional[Dict[Text, Union[FingerprintStatus, Any]]] = None


def _dry_run_result(
    fingerprint_results: Dict[Text, Union[FingerprintStatus, Any]],
    force_full_training: bool,
) -> TrainingResult:
    """Returns a dry run result.

    Args:
        fingerprint_results: A result of fingerprint run..
        force_full_training: Whether the user used the `--force` flag to enforce a
            full retraining of the model.

    Returns:
        Result containing the return code and the fingerprint results.
    """
    if force_full_training:
        structlogger.warn(
            "model_training.force_full_training", event_info="The training was forced."
        )
        return TrainingResult(
            code=CODE_FORCED_TRAINING, dry_run_results=fingerprint_results
        )

    training_required = any(
        isinstance(result, FingerprintStatus) and not result.is_hit
        for result in fingerprint_results.values()
    )

    if training_required:
        structlogger.warn(
            "model_training.training_required",
            event_info="The model needs to be retrained.",
        )
        return TrainingResult(
            code=CODE_NEEDS_TO_BE_RETRAINED, dry_run_results=fingerprint_results
        )

    structlogger.info(
        "model_training.no_training_required",
        event_info=(
            "No training of components required "
            "(the responses might still need updating!)."
        ),
    )
    return TrainingResult(
        code=CODE_NO_NEED_TO_TRAIN, dry_run_results=fingerprint_results
    )


def get_unresolved_slots(domain: Domain, stories: StoryGraph) -> List[Text]:
    """Returns a list of unresolved slots.

    Args:
        domain: The domain.
        stories: The story graph.

    Returns:
        A list of unresolved slots.
    """
    return list(
        set(
            evnt.key
            for step in stories.story_steps
            for evnt in step.events
            if isinstance(evnt, SlotSet)
        )
        - set(slot.name for slot in domain.slots)
    )


def _check_unresolved_slots(domain: Domain, stories: StoryGraph) -> None:
    """Checks if there are any unresolved slots.

    Args:
        domain: The domain.
        stories: The story graph.

    Raises:
        `Sys exit` if there are any unresolved slots.

    Returns:
        `None` if there are no unresolved slots.
    """
    unresolved_slots = get_unresolved_slots(domain, stories)
    if unresolved_slots:
        structlogger.error(
            "model.training.check_unresolved_slots.not_in_domain",
            slots=unresolved_slots,
            event_info=(
                f"Unresolved slots found in stories/rules🚨 \n"
                f'Tried to set slots "{unresolved_slots}" that are not present in'
                f"your domain.\n Check whether they need to be added to the domain or "
                f"whether there is a spelling error."
            ),
        )
        sys.exit(1)


async def train(
    domain: Text,
    config: Text,
    training_files: Optional[Union[Text, List[Text]]],
    output: Text = rasa.shared.constants.DEFAULT_MODELS_PATH,
    dry_run: bool = False,
    force_training: bool = False,
    fixed_model_name: Optional[Text] = None,
    persist_nlu_training_data: bool = False,
    core_additional_arguments: Optional[Dict] = None,
    nlu_additional_arguments: Optional[Dict] = None,
    model_to_finetune: Optional[Text] = None,
    finetuning_epoch_fraction: float = 1.0,
    remote_storage: Optional[StorageType] = None,
    file_importer: Optional[TrainingDataImporter] = None,
    keep_local_model_copy: bool = False,
) -> TrainingResult:
    """Trains a Rasa model (Core and NLU).

    Args:
        domain: Path to the domain file.
        config: Path to the config file.
        training_files: List of paths to training data files.
        output: Output directory for the trained model.
        dry_run: If `True` then no training will be done, and the information about
            whether the training needs to be done will be printed.
        force_training: If `True` retrain model even if data has not changed.
        fixed_model_name: Name of model to be stored.
        persist_nlu_training_data: `True` if the NLU training data should be persisted
            with the model.
        core_additional_arguments: Additional training parameters for core training.
        nlu_additional_arguments: Additional training parameters forwarded to training
            method of each NLU component.
        model_to_finetune: Optional path to a model which should be finetuned or
            a directory in case the latest trained model should be used.
        finetuning_epoch_fraction: The fraction currently specified training epochs
            in the model configuration which should be used for finetuning.
        remote_storage: Optional name of the remote storage to
            use for storing the model.
        file_importer: Instance of `TrainingDataImporter` to use for training.
            If it is not provided, a new instance will be created.
        keep_local_model_copy: If `True` the model will be stored locally even if
            remote storage is configured.

    Returns:
        An instance of `TrainingResult`.
    """
    if not file_importer:
        file_importer = TrainingDataImporter.load_from_config(
            config, domain, training_files, core_additional_arguments
        )

    stories = file_importer.get_stories()
    flows = file_importer.get_flows()
    nlu_data = file_importer.get_nlu_data()

    training_type = TrainingType.BOTH

    if nlu_data.has_e2e_examples():
        rasa.shared.utils.common.mark_as_experimental_feature("end-to-end training")
        training_type = TrainingType.END_TO_END

    if stories.is_empty() and nlu_data.contains_no_pure_nlu_data() and flows.is_empty():
        structlogger.error(
            "model_training.train.no_training_data_found",
            event_info=(
                "No training data given. Please provide "
                "stories, flows or NLU data in "
                "order to train a Rasa model using the "
                "'--data' argument."
            ),
        )
        return TrainingResult(code=1)

    domain_object = file_importer.get_domain()
    if domain_object.is_empty():
        structlogger.warn(
            "model_training.train.domain_not_found",
            event_info=(
                "Core training was skipped because no "
                "valid domain file was found. Only an "
                "NLU-model was created. Please specify "
                "a valid domain using the '--domain' "
                "argument or check if the provided domain "
                "file exists."
            ),
        )
        training_type = TrainingType.NLU

    elif stories.is_empty() and flows.is_empty():
        structlogger.warn(
            "model_training.train.flows_and_stories_not_found",
            event_info=(
                "No stories or flows present. Just a " "Rasa NLU model will be trained."
            ),
        )
        training_type = TrainingType.NLU

    # We will train nlu if there are any nlu example, including from e2e stories.
    elif (
        nlu_data.contains_no_pure_nlu_data()
        and not nlu_data.has_e2e_examples()
        and flows.is_empty()
    ):
        structlogger.warn(
            "model_training.train.nlu_data_not_found",
            event_info="No NLU data present. No NLU model will be trained.",
        )
        training_type = TrainingType.CORE

    _check_unresolved_slots(domain_object, stories)

    with telemetry.track_model_training(file_importer, model_type="rasa"):
        return await _train_graph(
            file_importer,
            training_type=training_type,
            output_path=output,
            fixed_model_name=fixed_model_name,
            model_to_finetune=model_to_finetune,
            force_full_training=force_training,
            persist_nlu_training_data=persist_nlu_training_data,
            finetuning_epoch_fraction=finetuning_epoch_fraction,
            dry_run=dry_run,
            remote_storage=remote_storage,
            keep_local_model_copy=keep_local_model_copy,
            **(core_additional_arguments or {}),
            **(nlu_additional_arguments or {}),
        )


async def _train_graph(
    file_importer: TrainingDataImporter,
    training_type: TrainingType,
    output_path: Text,
    fixed_model_name: Text,
    model_to_finetune: Optional[Union[Text, Path]] = None,
    force_full_training: bool = False,
    dry_run: bool = False,
    remote_storage: Optional[StorageType] = None,
    keep_local_model_copy: bool = False,
    **kwargs: Any,
) -> TrainingResult:
    if model_to_finetune:
        model_to_finetune = rasa.model.get_model_for_finetuning(model_to_finetune)
        if not model_to_finetune:
            structlogger.error(
                "model_training.train.finetuning_model_not_found",
                event_info=(
                    f"No model for finetuning found. Please make sure to either "
                    f"specify a path to a previous model or to have a finetunable "
                    f"model within the directory '{output_path}'."
                ),
            )
            sys.exit(1)

        rasa.shared.utils.common.mark_as_experimental_feature(
            "Incremental Training feature"
        )

    is_finetuning = model_to_finetune is not None

    config = file_importer.get_config()
    recipe = Recipe.recipe_for_name(config.get(CONFIG_RECIPE_KEY))
    config, _missing_keys, _configured_keys = recipe.auto_configure(
        file_importer.get_config_file_for_auto_config(),
        config,
        training_type,
    )
    flows = file_importer.get_flows()
    domain = file_importer.get_domain()
    model_configuration = recipe.graph_config_for_recipe(
        config,
        kwargs,
        training_type=training_type,
        is_finetuning=is_finetuning,
    )
    rasa.engine.validation.validate(model_configuration)
    rasa.engine.validation.validate_coexistance_routing_setup(
        domain, model_configuration, flows
    )
    rasa.engine.validation.validate_model_group_configuration_setup()
    rasa.engine.validation.validate_model_client_configuration_setup_during_training_time(
        config
    )
    rasa.engine.validation.validate_flow_component_dependencies(
        flows, model_configuration
    )
    rasa.engine.validation.validate_command_generator_setup(model_configuration)

    tempdir_name = rasa.utils.common.get_temp_dir_name()
    # Use `TempDirectoryPath` instead of `tempfile.TemporaryDirectory` as this
    # leads to errors on Windows when the context manager tries to delete an
    # already deleted temporary directory (e.g. https://bugs.python.org/issue29982)
    with rasa.utils.common.TempDirectoryPath(tempdir_name) as temp_model_dir:
        model_storage = _create_model_storage(
            is_finetuning, model_to_finetune, Path(temp_model_dir)
        )
        cache = LocalTrainingCache()
        trainer = GraphTrainer(model_storage, cache, DaskGraphRunner)

        if dry_run:
            fingerprint_status = await trainer.fingerprint(
                model_configuration.train_schema, file_importer
            )
            return _dry_run_result(fingerprint_status, force_full_training)

        model_name = determine_model_name(fixed_model_name, training_type)
        full_model_path = Path(output_path, model_name)

        await trainer.train(
            model_configuration,
            file_importer,
            full_model_path,
            force_retraining=force_full_training,
            is_finetuning=is_finetuning,
        )
        if remote_storage:
            push_model_to_remote_storage(full_model_path, remote_storage)
            if not keep_local_model_copy:
                full_model_path.unlink()
            structlogger.info(
                "model_training.train.finished_training",
                event_info=(
                    f"Your Rasa model {model_name} is trained "
                    f"and saved at remote storage provider '{remote_storage}'."
                ),
            )
        else:
            structlogger.info(
                "model_training.train.finished_training",
                event_info=(
                    f"Your Rasa model is trained and saved at '{full_model_path}'."
                ),
            )

        return TrainingResult(str(full_model_path), 0)


def _create_model_storage(
    is_finetuning: bool, model_to_finetune: Optional[Path], temp_model_dir: Path
) -> ModelStorage:
    if is_finetuning:
        model_storage, _ = LocalModelStorage.from_model_archive(
            temp_model_dir, model_to_finetune
        )
    else:
        model_storage = LocalModelStorage(temp_model_dir)

    return model_storage


def generate_random_model_name() -> str:
    time_format = "%Y%m%d-%H%M%S"
    return f"{time.strftime(time_format)}-{randomname.get_name()}"


def determine_model_name(
    fixed_model_name: Optional[Text], training_type: TrainingType
) -> str:
    if fixed_model_name:
        if not fixed_model_name.endswith(".tar.gz"):
            return f"{fixed_model_name}.tar.gz"
        return fixed_model_name

    prefix = ""
    if training_type in [TrainingType.CORE, TrainingType.NLU]:
        prefix = f"{training_type.model_type}-"

    return f"{prefix}{generate_random_model_name()}.tar.gz"


async def train_core(
    domain: Union[Domain, Text],
    config: Text,
    stories: Text,
    output: Text,
    fixed_model_name: Optional[Text] = None,
    additional_arguments: Optional[Dict] = None,
    model_to_finetune: Optional[Text] = None,
    finetuning_epoch_fraction: float = 1.0,
    keep_local_model_copy: bool = False,
) -> Optional[Text]:
    """Trains a Core model.

    Args:
        domain: Path to the domain file.
        config: Path to the config file for Core.
        stories: Path to the Core training data.
        output: Output path.
        fixed_model_name: Name of model to be stored.
        additional_arguments: Additional training parameters.
        model_to_finetune: Optional path to a model which should be finetuned or
            a directory in case the latest trained model should be used.
        finetuning_epoch_fraction: The fraction currently specified training epochs
            in the model configuration which should be used for finetuning.
        keep_local_model_copy: If `True` the model will be stored locally even if
            remote storage is configured.

    Returns:
        Path to the model archive.

    """
    file_importer = TrainingDataImporter.load_core_importer_from_config(
        config, domain, [stories], additional_arguments
    )
    stories_data = file_importer.get_stories()
    nlu_data = file_importer.get_nlu_data()
    domain = file_importer.get_domain()

    if nlu_data.has_e2e_examples():
        structlogger.error(
            "model_training.train_core.e2e_stories_found",
            event_info=(
                "Stories file contains e2e stories. "
                "Please train using `rasa train` so that "
                "the NLU model is also trained."
            ),
        )
        return None

    if domain.is_empty():
        structlogger.error(
            "model_training.train_core.domain_not_found",
            event_info=(
                "Core training was skipped because no valid "
                "domain file was found. Please specify a valid "
                "domain using '--domain' argument or check "
                "if the provided domain file exists."
            ),
        )
        return None

    if not stories_data:
        structlogger.error(
            "model_training.train_core.stories_not_found",
            event_info=(
                "No stories given. Please provide stories in order to "
                "train a Rasa Core model using the '--stories' argument."
            ),
        )
        return None

    _check_unresolved_slots(domain, stories_data)

    return (
        await _train_graph(
            file_importer,
            training_type=TrainingType.CORE,
            output_path=output,
            model_to_finetune=model_to_finetune,
            fixed_model_name=fixed_model_name,
            finetuning_epoch_fraction=finetuning_epoch_fraction,
            keep_local_model_copy=keep_local_model_copy,
            **(additional_arguments or {}),
        )
    ).model


async def train_nlu(
    config: Text,
    nlu_data: Optional[Text],
    output: Text,
    fixed_model_name: Optional[Text] = None,
    persist_nlu_training_data: bool = False,
    additional_arguments: Optional[Dict] = None,
    domain: Optional[Union[Domain, Text]] = None,
    model_to_finetune: Optional[Text] = None,
    finetuning_epoch_fraction: float = 1.0,
    keep_local_model_copy: bool = False,
) -> Optional[Text]:
    """Trains an NLU model.

    Args:
        config: Path to the config file for NLU.
        nlu_data: Path to the NLU training data.
        output: Output path.
        fixed_model_name: Name of the model to be stored.
        persist_nlu_training_data: `True` if the NLU training data should be persisted
                                   with the model.
        additional_arguments: Additional training parameters which will be passed to
                              the `train` method of each component.
        domain: Path to the optional domain file/Domain object.
        model_to_finetune: Optional path to a model which should be finetuned or
            a directory in case the latest trained model should be used.
        finetuning_epoch_fraction: The fraction currently specified training epochs
            in the model configuration which should be used for finetuning.
        keep_local_model_copy: If `True` the model will be stored locally even if
            remote storage is configured.

    Returns:
        Path to the model archive.
    """
    if not nlu_data:
        structlogger.error(
            "model_training.train_nlu.nlu_data_not_found",
            event_info=(
                "No NLU data given. Please provide NLU data in order to train "
                "a Rasa NLU model using the '--nlu' argument."
            ),
        )
        return None

    # training NLU only hence the training files still have to be selected
    file_importer = TrainingDataImporter.load_nlu_importer_from_config(
        config, domain, training_data_paths=[nlu_data], args=additional_arguments
    )

    training_data = file_importer.get_nlu_data()
    if training_data.contains_no_pure_nlu_data():
        structlogger.error(
            "model_training.train_nlu.nlu_data_invalid",
            path=nlu_data,
            event_info=(
                f"Path '{nlu_data}' doesn't contain valid NLU data in it. "
                f"Please verify the data format. "
                f"The NLU model training will be skipped now."
            ),
        )
        return None

    return (
        await _train_graph(
            file_importer,
            training_type=TrainingType.NLU,
            output_path=output,
            model_to_finetune=model_to_finetune,
            fixed_model_name=fixed_model_name,
            finetuning_epoch_fraction=finetuning_epoch_fraction,
            persist_nlu_training_data=persist_nlu_training_data,
            keep_local_model_copy=keep_local_model_copy,
            **(additional_arguments or {}),
        )
    ).model


def push_model_to_remote_storage(model_path: Path, remote_storage: StorageType) -> None:
    """Push model to remote storage."""
    from rasa.core.persistor import get_persistor

    persistor = get_persistor(remote_storage)

    if persistor is not None:
        persistor.persist(str(model_path))

    else:
        raise RasaException(
            f"Persistor not found for remote storage: '{remote_storage}'."
        )



from enum import Enum
from typing import Optional

import structlog


class CommandSyntaxVersion(Enum):
    """Defines different syntax versions for commands."""

    v1 = "v1"
    v2 = "v2"


structlogger = structlog.get_logger()


class CommandSyntaxManager:
    """A class to manage the command syntax version. It is used to set and get the
    command syntax version. This class provides a way to introduce new syntax versions
    for commands in the future. Hence, it is for internal use only.
    """

    _version = None  # Directly store the version as a class attribute

    @classmethod
    def set_syntax_version(cls, version: CommandSyntaxVersion) -> None:
        """Sets the command syntax version on the class itself.
        This method is called only once at the time of LLMCommandGenerator
        initialization to set the command syntax version, which ensures that the command
        syntax version remains consistent throughout the lifetime of the generator.
        """
        if cls._version:
            structlogger.warn(
                "command_syntax_manager.syntax_version_already_set",
                event_info=(
                    f"The command syntax version has already been set. Overwriting "
                    f"the existing version with the new version - {version}."
                ),
            )
        cls._version = version

    @classmethod
    def get_syntax_version(cls) -> Optional[CommandSyntaxVersion]:
        """Fetches the stored command syntax version."""
        return cls._version

    @staticmethod
    def get_default_syntax_version() -> CommandSyntaxVersion:
        """Returns the default command syntax version."""
        return CommandSyntaxVersion.v1

    @classmethod
    def reset_syntax_version(cls) -> None:
        """Resets the command syntax version. Implemented for use in testing."""
        cls._version = None



from __future__ import annotations

import copy
import re
from dataclasses import dataclass
from typing import Any, Dict, List

import structlog

from rasa.dialogue_understanding.commands.command import Command
from rasa.dialogue_understanding.commands.command_syntax_manager import (
    CommandSyntaxManager,
    CommandSyntaxVersion,
)
from rasa.dialogue_understanding.patterns.cancel import CancelPatternFlowStackFrame
from rasa.dialogue_understanding.patterns.clarify import ClarifyPatternFlowStackFrame
from rasa.dialogue_understanding.stack.dialogue_stack import DialogueStack
from rasa.dialogue_understanding.stack.frames import UserFlowStackFrame
from rasa.dialogue_understanding.stack.frames.flow_stack_frame import FlowStackFrameType
from rasa.dialogue_understanding.stack.utils import top_user_flow_frame
from rasa.shared.core.events import Event, FlowCancelled
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.trackers import DialogueStateTracker

structlogger = structlog.get_logger()


@dataclass
class CancelFlowCommand(Command):
    """A command to cancel the current flow."""

    @classmethod
    def command(cls) -> str:
        """Returns the command type."""
        return "cancel flow"

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> CancelFlowCommand:
        """Converts the dictionary to a command.

        Returns:
            The converted dictionary.
        """
        return CancelFlowCommand()

    @staticmethod
    def select_canceled_frames(stack: DialogueStack) -> List[str]:
        """Selects the frames that were canceled.

        Args:
            dialogue_stack: The dialogue stack.
            current_flow: The current flow.

        Returns:
        The frames that were canceled.
        """
        canceled_frames = []
        # we need to go through the original stack dump in reverse order
        # to find the frames that were canceled. we cancel everything from
        # the top of the stack until we hit the user flow that was canceled.
        # this will also cancel any patterns put on top of that user flow,
        # e.g. corrections.
        for frame in reversed(stack.frames):
            canceled_frames.append(frame.frame_id)
            if (
                isinstance(frame, UserFlowStackFrame)
                and frame.frame_type != FlowStackFrameType.CALL
            ):
                return canceled_frames
        else:
            # we should never get here as we should always find the user flow
            # that was canceled.
            raise ValueError(
                f"Could not find a user flow frame to cancel. "
                f"Current stack: {stack}."
            )

    def run_command_on_tracker(
        self,
        tracker: DialogueStateTracker,
        all_flows: FlowsList,
        original_tracker: DialogueStateTracker,
    ) -> List[Event]:
        """Runs the command on the tracker.

        Args:
            tracker: The tracker to run the command on.
            all_flows: All flows in the assistant.
            original_tracker: The tracker before any command was executed.

        Returns:
            The events to apply to the tracker.
        """
        stack = tracker.stack
        original_stack = original_tracker.stack

        applied_events: List[Event] = []
        # capture the top frame before we push new frames onto the stack
        initial_top_frame = stack.top()
        user_frame = top_user_flow_frame(original_stack)
        current_flow = user_frame.flow(all_flows) if user_frame else None

        if not current_flow:
            structlogger.debug(
                "command_executor.skip_cancel_flow.no_active_flow", command=self
            )
            return []

        # we pass in the original dialogue stack (before any of the currently
        # predicted commands were applied) to make sure we don't cancel any
        # frames that were added by the currently predicted commands.
        canceled_frames = self.select_canceled_frames(original_stack)

        stack.push(
            CancelPatternFlowStackFrame(
                canceled_name=current_flow.readable_name(
                    language=tracker.current_language
                ),
                canceled_frames=canceled_frames,
            )
        )

        if user_frame:
            applied_events.append(FlowCancelled(user_frame.flow_id, user_frame.step_id))

        if initial_top_frame and isinstance(
            initial_top_frame, ClarifyPatternFlowStackFrame
        ):
            structlogger.debug(
                "command_executor.cancel_flow.cancel_clarification_options",
                clarification_options=initial_top_frame.clarification_options,
            )
            applied_events += cancel_all_pending_clarification_options(
                initial_top_frame,
                original_stack,
                canceled_frames,
                all_flows,
                stack,
            )

        return applied_events + tracker.create_stack_updated_events(stack)

    def __hash__(self) -> int:
        return hash(self.command())

    def __eq__(self, other: object) -> bool:
        return isinstance(other, CancelFlowCommand)

    def to_dsl(self) -> str:
        """Converts the command to a DSL string."""
        mapper = {
            CommandSyntaxVersion.v1: "CancelFlow()",
            CommandSyntaxVersion.v2: "cancel flow",
        }
        return mapper.get(
            CommandSyntaxManager.get_syntax_version(),
            mapper[CommandSyntaxManager.get_default_syntax_version()],
        )

    @classmethod
    def from_dsl(cls, match: re.Match, **kwargs: Any) -> CancelFlowCommand:
        """Converts a DSL string to a command."""
        return CancelFlowCommand()

    @staticmethod
    def regex_pattern() -> str:
        mapper = {
            CommandSyntaxVersion.v1: r"CancelFlow\(\)",
            CommandSyntaxVersion.v2: r"^[^\w]*cancel flow$",
        }
        return mapper.get(
            CommandSyntaxManager.get_syntax_version(),
            mapper[CommandSyntaxManager.get_default_syntax_version()],
        )


def cancel_all_pending_clarification_options(
    initial_top_frame: ClarifyPatternFlowStackFrame,
    original_stack: DialogueStack,
    canceled_frames: List[str],
    all_flows: FlowsList,
    stack: DialogueStack,
) -> List[FlowCancelled]:
    """Cancel all pending clarification options.

    This is a special case when the assistant asks the user to clarify
    which pending digression flow to start after the completion of an active flow.
    If the user chooses to cancel all options, this function takes care of
    updating the stack by removing all pending flow stack frames
    listed as clarification options.
    """
    clarification_names = set(initial_top_frame.names)
    to_be_canceled_frames = []
    applied_events = []
    for frame in reversed(original_stack.frames):
        if frame.frame_id in canceled_frames:
            continue

        to_be_canceled_frames.append(frame.frame_id)
        if isinstance(frame, UserFlowStackFrame):
            readable_flow_name = frame.flow(all_flows).readable_name()
            if readable_flow_name in clarification_names:
                stack.push(
                    CancelPatternFlowStackFrame(
                        canceled_name=readable_flow_name,
                        canceled_frames=copy.deepcopy(to_be_canceled_frames),
                    )
                )
                applied_events.append(FlowCancelled(frame.flow_id, frame.step_id))
                to_be_canceled_frames.clear()

    return applied_events



from __future__ import annotations

from typing import List, Text

from rasa.shared.engine.caching import get_local_cache_location

DOCS_BASE_URL = "https://rasa.com/docs/rasa-pro"
DOCS_URL_CONCEPTS = DOCS_BASE_URL + "/concepts"
DOCS_URL_NLU_BASED = DOCS_BASE_URL + "/nlu-based-assistants"

# Docs URLs for CALM assistants
DOCS_URL_SLOTS = DOCS_URL_CONCEPTS + "/domain#slots"
DOCS_URL_RESPONSES = DOCS_URL_CONCEPTS + "/responses"
DOCS_URL_POLICIES = DOCS_URL_CONCEPTS + "/policies/policy-overview"
DOCS_URL_ACTIONS = DOCS_URL_CONCEPTS + "/actions"
DOCS_URL_DEFAULT_ACTIONS = DOCS_URL_CONCEPTS + "/default-actions"
DOCS_URL_COMPONENTS = DOCS_URL_CONCEPTS + "/components/overview"
DOCS_URL_GRAPH_COMPONENTS = DOCS_URL_CONCEPTS + "/components/custom-graph-components"
DOCS_URL_GRAPH_RECIPE = DOCS_URL_CONCEPTS + "/components/graph-recipe"
DOCS_URL_CATEGORICAL_SLOTS = DOCS_URL_CONCEPTS + "/domain#categorical-slot"
DOCS_URL_DOMAIN = DOCS_URL_CONCEPTS + "/domain"
DOCS_URL_FLOWS = DOCS_URL_CONCEPTS + "/flows"

# Docs URLs for NLU-based assistants
DOCS_URL_TRAINING_DATA = DOCS_URL_NLU_BASED + "/training-data-format"
DOCS_URL_TRAINING_DATA_NLU = DOCS_URL_TRAINING_DATA + "#nlu-training-data"
DOCS_URL_DOMAINS = DOCS_URL_NLU_BASED + "/domain"
DOCS_URL_NLU_BASED_SLOTS = DOCS_URL_DOMAINS + "#slots"
DOCS_URL_INTENTS = DOCS_URL_DOMAINS + "#intents"
DOCS_URL_ENTITIES = DOCS_URL_DOMAINS + "#entities"
DOCS_URL_STORIES = DOCS_URL_NLU_BASED + "/stories"
DOCS_URL_RULES = DOCS_URL_NLU_BASED + "/rules"
DOCS_URL_FORMS = DOCS_URL_NLU_BASED + "/forms"
DOCS_URL_PIPELINE = DOCS_URL_NLU_BASED + "/model-configuration"
DOCS_URL_NLU_BASED_POLICIES = DOCS_URL_NLU_BASED + "/policies"

# Other docs URLs
DOCS_URL_MARKERS = DOCS_BASE_URL + "/operating/analytics/realtime-markers"
DOCS_URL_CONNECTORS = DOCS_BASE_URL + "/connectors/messaging-and-voice-channels"
DOCS_URL_CONNECTORS_SLACK = DOCS_BASE_URL + "/connectors/slack"
DOCS_URL_EVENT_BROKERS = DOCS_BASE_URL + "/production/event-brokers"
DOCS_URL_PIKA_EVENT_BROKER = DOCS_URL_EVENT_BROKERS + "#pika-event-broker"
DOCS_URL_TRACKER_STORES = DOCS_BASE_URL + "/production/tracker-stores"
DOCS_URL_MIGRATION_GUIDE = DOCS_BASE_URL + "/migration-guide"
DOCS_URL_TELEMETRY = DOCS_BASE_URL + "/telemetry/telemetry"

INTENT_MESSAGE_PREFIX = "/"

PACKAGE_NAME = "rasa"
NEXT_MAJOR_VERSION_FOR_DEPRECATIONS = "4.0.0"

MODEL_CONFIG_SCHEMA_FILE = "shared/utils/schemas/model_config.yml"
CONFIG_SCHEMA_FILE = "shared/utils/schemas/config.yml"
RESPONSES_SCHEMA_FILE = "shared/nlu/training_data/schemas/responses.yml"
SCHEMA_EXTENSIONS_FILE = "shared/utils/pykwalify_extensions.py"
ASSERTIONS_SCHEMA_FILE = "e2e_test/assertions_schema.yml"
ASSERTIONS_SCHEMA_EXTENSIONS_FILE = "e2e_test/pykwalify_extensions.py"
LATEST_TRAINING_DATA_FORMAT_VERSION = "3.1"

DOMAIN_SCHEMA_FILE = "shared/utils/schemas/domain.yml"

DEFAULT_SESSION_EXPIRATION_TIME_IN_MINUTES = 60
DEFAULT_CARRY_OVER_SLOTS_TO_NEW_SESSION = True

DEFAULT_NLU_FALLBACK_INTENT_NAME = "nlu_fallback"

DEFAULT_E2E_TESTS_PATH = "."
TEST_STORIES_FILE_PREFIX = "test_"

LOG_LEVEL_NAME_TO_LEVEL = {
    "CRITICAL": 50,
    "ERROR": 40,
    "WARNING": 30,
    "INFO": 20,
    "DEBUG": 10,
    "NOTSET": 0,
}

DEFAULT_LOG_LEVEL = "INFO"
ENV_LOG_LEVEL = "LOG_LEVEL"
DEFAULT_LOG_LEVEL_LLM = "DEBUG"
ENV_LOG_LEVEL_LLM = "LOG_LEVEL_LLM"
ENV_LOG_LEVEL_LLM_MODULE_NAMES = {
    "LLMCommandGenerator": "LOG_LEVEL_LLM_COMMAND_GENERATOR",
    "SingleStepLLMCommandGenerator": "LOG_LEVEL_LLM_COMMAND_GENERATOR",
    "CompactLLMCommandGenerator": "LOG_LEVEL_LLM_COMMAND_GENERATOR",
    "MultiStepLLMCommandGenerator": "LOG_LEVEL_LLM_COMMAND_GENERATOR",
    "EnterpriseSearchPolicy": "LOG_LEVEL_LLM_ENTERPRISE_SEARCH",
    "IntentlessPolicy": "LOG_LEVEL_LLM_INTENTLESS_POLICY",
    "ContextualResponseRephraser": "LOG_LEVEL_LLM_REPHRASER",
    "NLUCommandAdapter": "LOG_LEVEL_NLU_COMMAND_ADAPTER",
    "LLMBasedRouter": "LOG_LEVEL_LLM_BASED_ROUTER",
}
TCP_PROTOCOL = "TCP"

DEFAULT_SENDER_ID = "default"
UTTER_PREFIX = "utter_"
UTTER_ASK_PREFIX = "utter_ask_"
ACTION_ASK_PREFIX = "action_ask_"
FLOW_PREFIX = "flow_"

ASSISTANT_ID_KEY = "assistant_id"
ASSISTANT_ID_DEFAULT_VALUE = "placeholder_default"

CONFIG_MANDATORY_COMMON_KEYS = [ASSISTANT_ID_KEY]
CONFIG_NAME_KEY = "name"
CONFIG_POLICIES_KEY = "policies"
CONFIG_PIPELINE_KEY = "pipeline"
CONFIG_LANGUAGE_KEY = "language"
CONFIG_ADDITIONAL_LANGUAGES_KEY = "additional_languages"
CONFIG_RECIPE_KEY = "recipe"
CONFIG_LLM_KEY = "llm"
CONFIG_MODEL_NAME_KEY = "model_name"
CONFIG_TRAIN_SCHEMA = "train_schema"
CONFIG_PREDICT_SCHEMA = "predict_schema"
CONFIG_AUTOCONFIGURABLE_KEYS_CORE = [CONFIG_POLICIES_KEY]
CONFIG_AUTOCONFIGURABLE_KEYS_NLU = [CONFIG_PIPELINE_KEY]
CONFIG_AUTOCONFIGURABLE_KEYS = (
    CONFIG_AUTOCONFIGURABLE_KEYS_CORE + CONFIG_AUTOCONFIGURABLE_KEYS_NLU
)
CONFIG_KEYS_CORE = [CONFIG_POLICIES_KEY] + CONFIG_MANDATORY_COMMON_KEYS
CONFIG_KEYS_NLU = [
    CONFIG_LANGUAGE_KEY,
    CONFIG_PIPELINE_KEY,
] + CONFIG_MANDATORY_COMMON_KEYS
CONFIG_KEYS = CONFIG_KEYS_CORE + CONFIG_KEYS_NLU
CONFIG_MANDATORY_KEYS_CORE: List[Text] = [] + CONFIG_MANDATORY_COMMON_KEYS
CONFIG_MANDATORY_KEYS_NLU = [CONFIG_LANGUAGE_KEY] + CONFIG_MANDATORY_COMMON_KEYS
CONFIG_MANDATORY_KEYS = CONFIG_MANDATORY_KEYS_CORE + CONFIG_MANDATORY_KEYS_NLU

# Keys related to Forms (in the Domain)
REQUIRED_SLOTS_KEY = "required_slots"
IGNORED_INTENTS = "ignored_intents"

# Constants for default Rasa Pro project layout
DEFAULT_ENDPOINTS_PATH = "endpoints.yml"
DEFAULT_CREDENTIALS_PATH = "credentials.yml"
DEFAULT_CONFIG_PATH = "config.yml"
DEFAULT_DOMAIN_PATHS = ["domain.yml", "domain"]
DEFAULT_DOMAIN_PATH = DEFAULT_DOMAIN_PATHS[0]
DEFAULT_ACTIONS_PATH = "actions"
DEFAULT_MODELS_PATH = "models"
DEFAULT_CONVERTED_DATA_PATH = "converted_data"
DEFAULT_DATA_PATH = "data"
DEFAULT_RESULTS_PATH = "results"
DEFAULT_NLU_RESULTS_PATH = "nlu_comparison_results"
DEFAULT_CORE_SUBDIRECTORY_NAME = "core"
DEFAULT_NLU_SUBDIRECTORY_NAME = "nlu"
DEFAULT_CONVERSATION_TEST_PATH = "tests"
DEFAULT_MARKERS_PATH = "markers"
DEFAULT_MARKERS_CONFIG_PATH = "markers/config"
DEFAULT_MARKERS_OUTPUT_PATH = "markers/output"
DEFAULT_MARKERS_STATS_PATH = "markers/stats"

DEFAULT_PROMPT_PACKAGE_NAME = "rasa.dialogue_understanding.generator.prompt_templates"

DIAGNOSTIC_DATA = "diagnostic_data"

RESPONSE_CONDITION = "condition"
CHANNEL = "channel"

API_KEY = "api_key"

AZURE_API_KEY_ENV_VAR = "AZURE_API_KEY"
AZURE_AD_TOKEN_ENV_VAR = "AZURE_AD_TOKEN"
AZURE_API_BASE_ENV_VAR = "AZURE_API_BASE"
AZURE_API_VERSION_ENV_VAR = "AZURE_API_VERSION"
AZURE_API_TYPE_ENV_VAR = "AZURE_API_TYPE"
AZURE_AD_SCOPES_ENV_VAR = "AZURE_AD_SCOPES"
AZURE_SPEECH_API_KEY_ENV_VAR = "AZURE_SPEECH_API_KEY"

DEEPGRAM_API_KEY_ENV_VAR = "DEEPGRAM_API_KEY"
CARTESIA_API_KEY_ENV_VAR = "CARTESIA_API_KEY"

OPENAI_API_KEY_ENV_VAR = "OPENAI_API_KEY"
OPENAI_API_TYPE_ENV_VAR = "OPENAI_API_TYPE"
OPENAI_API_VERSION_ENV_VAR = "OPENAI_API_VERSION"
OPENAI_API_BASE_ENV_VAR = "OPENAI_API_BASE"

OPENAI_API_BASE_CONFIG_KEY = "openai_api_base"
OPENAI_API_TYPE_CONFIG_KEY = "openai_api_type"
OPENAI_API_VERSION_CONFIG_KEY = "openai_api_version"

AWS_BEDROCK_PROVIDER = "bedrock"
AWS_SAGEMAKER_PROVIDER = "sagemaker"

API_BASE_CONFIG_KEY = "api_base"
API_TYPE_CONFIG_KEY = "api_type"
API_VERSION_CONFIG_KEY = "api_version"
LANGCHAIN_TYPE_CONFIG_KEY = "_type"
RASA_TYPE_CONFIG_KEY = "type"
PROVIDER_CONFIG_KEY = "provider"

REQUEST_TIMEOUT_CONFIG_KEY = "request_timeout"  # deprecated
TIMEOUT_CONFIG_KEY = "timeout"

DEPLOYMENT_NAME_CONFIG_KEY = "deployment_name"
DEPLOYMENT_CONFIG_KEY = "deployment"
EMBEDDINGS_CONFIG_KEY = "embeddings"
ENGINE_CONFIG_KEY = "engine"
LLM_CONFIG_KEY = "llm"
MODEL_CONFIG_KEY = "model"
MODEL_NAME_CONFIG_KEY = "model_name"
PROMPT_CONFIG_KEY = "prompt"
PROMPT_TEMPLATE_CONFIG_KEY = "prompt_template"
STREAM_CONFIG_KEY = "stream"
N_REPHRASES_CONFIG_KEY = "n"
USE_CHAT_COMPLETIONS_ENDPOINT_CONFIG_KEY = "use_chat_completions_endpoint"

ROUTER_CONFIG_KEY = "router"
ROUTING_STRATEGY_CONFIG_KEY = "routing_strategy"
REDIS_HOST_CONFIG_KEY = "redis_host"
ROUTING_STRATEGIES_REQUIRING_REDIS_CACHE = [
    "cost-based-routing",
    "usage-based-routing",
]
ROUTING_STRATEGIES_NOT_REQUIRING_CACHE = [
    "latency-based-routing",
    "least-busy",
    "simple-shuffle",
]
VALID_ROUTING_STRATEGIES = (
    ROUTING_STRATEGIES_REQUIRING_REDIS_CACHE + ROUTING_STRATEGIES_NOT_REQUIRING_CACHE
)

MODELS_CONFIG_KEY = "models"
MODEL_GROUPS_CONFIG_KEY = "model_groups"
MODEL_GROUP_CONFIG_KEY = "model_group"
MODEL_GROUP_ID_CONFIG_KEY = "id"

EXTRA_PARAMETERS_KEY = "extra_parameters"
MODEL_GROUP_ID_KEY = "model_group_id"
MODEL_LIST_KEY = "model_list"
LITELLM_PARAMS_KEY = "litellm_params"

LLM_API_HEALTH_CHECK_ENV_VAR = "LLM_API_HEALTH_CHECK"
LLM_API_HEALTH_CHECK_DEFAULT_VALUE = "false"

AWS_REGION_NAME_CONFIG_KEY = "aws_region_name"
AWS_ACCESS_KEY_ID_CONFIG_KEY = "aws_access_key_id"
AWS_SECRET_ACCESS_KEY_CONFIG_KEY = "aws_secret_access_key"
AWS_SESSION_TOKEN_CONFIG_KEY = "aws_session_token"

AWS_ACCESS_KEY_ID_ENV_VAR = "AWS_ACCESS_KEY_ID"
AWS_SECRET_ACCESS_KEY_ENV_VAR = "AWS_SECRET_ACCESS_KEY"
AWS_REGION_NAME_ENV_VAR = "AWS_REGION_NAME"
AWS_SESSION_TOKEN_ENV_VAR = "AWS_SESSION_TOKEN"

HUGGINGFACE_MULTIPROCESS_CONFIG_KEY = "multi_process"
HUGGINGFACE_CACHE_FOLDER_CONFIG_KEY = "cache_folder"
HUGGINGFACE_SHOW_PROGRESS_CONFIG_KEY = "show_progress"
HUGGINGFACE_MODEL_KWARGS_CONFIG_KEY = "model_kwargs"
HUGGINGFACE_ENCODE_KWARGS_CONFIG_KEY = "encode_kwargs"
HUGGINGFACE_LOCAL_EMBEDDING_CACHING_FOLDER = (
    get_local_cache_location() / "huggingface_local_embeddings"
)

REQUESTS_CA_BUNDLE_ENV_VAR = "REQUESTS_CA_BUNDLE"
REQUESTS_SSL_CONTEXT_PURPOSE_ENV_VAR = "REQUESTS_SSL_CONTEXT_PURPOSE"
RASA_CA_BUNDLE_ENV_VAR = "RASA_CA_BUNDLE"  # used in verify
RASA_SSL_CERTIFICATE_ENV_VAR = "RASA_SSL_CERTIFICATE"  # used in cert (client side)
LITELLM_SSL_VERIFY_ENV_VAR = "SSL_VERIFY"
LITELLM_SSL_CERTIFICATE_ENV_VAR = "SSL_CERTIFICATE"

OPENAI_PROVIDER = "openai"
AZURE_OPENAI_PROVIDER = "azure"
SELF_HOSTED_PROVIDER = "self-hosted"
HUGGINGFACE_LOCAL_EMBEDDING_PROVIDER = "huggingface_local"
RASA_PROVIDER = "rasa"

SELF_HOSTED_VLLM_PREFIX = "hosted_vllm"
SELF_HOSTED_VLLM_API_KEY_ENV_VAR = "HOSTED_VLLM_API_KEY"

VALID_PROVIDERS_FOR_API_TYPE_CONFIG_KEY = [
    OPENAI_PROVIDER,
    AZURE_OPENAI_PROVIDER,
]

AZURE_API_TYPE = "azure"
OPENAI_API_TYPE = "openai"

RASA_DEFAULT_FLOW_PATTERN_PREFIX = "pattern_"
CONTEXT = "context"

RASA_PATTERN_INTERNAL_ERROR = "pattern_internal_error"
RASA_PATTERN_HUMAN_HANDOFF = "pattern_human_handoff"

RASA_INTERNAL_ERROR_PREFIX = "rasa_internal_error_"
RASA_PATTERN_INTERNAL_ERROR_DEFAULT = RASA_INTERNAL_ERROR_PREFIX + "default"
RASA_PATTERN_INTERNAL_ERROR_USER_INPUT_TOO_LONG = (
    RASA_INTERNAL_ERROR_PREFIX + "user_input_too_long"
)
RASA_PATTERN_INTERNAL_ERROR_USER_INPUT_EMPTY = (
    RASA_INTERNAL_ERROR_PREFIX + "user_input_empty"
)

RASA_PATTERN_CANNOT_HANDLE_PREFIX = "cannot_handle_"
RASA_PATTERN_CANNOT_HANDLE_DEFAULT = RASA_PATTERN_CANNOT_HANDLE_PREFIX + "default"
RASA_PATTERN_CANNOT_HANDLE_CHITCHAT = RASA_PATTERN_CANNOT_HANDLE_PREFIX + "chitchat"
RASA_PATTERN_CANNOT_HANDLE_NOT_SUPPORTED = (
    RASA_PATTERN_CANNOT_HANDLE_PREFIX + "not_supported"
)
RASA_PATTERN_CANNOT_HANDLE_INVALID_INTENT = (
    RASA_PATTERN_CANNOT_HANDLE_PREFIX + "invalid_intent"
)

ROUTE_TO_CALM_SLOT = "route_session_to_calm"

SENSITIVE_DATA = [
    API_KEY,
    AWS_ACCESS_KEY_ID_CONFIG_KEY,
    AWS_SECRET_ACCESS_KEY_CONFIG_KEY,
    AWS_SESSION_TOKEN_CONFIG_KEY,
]

# Used for key values in BotUtterance.data
TEXT = "text"
ELEMENTS = "elements"
QUICK_REPLIES = "quick_replies"
BUTTONS = "buttons"
ATTACHMENT = "attachment"
IMAGE = "image"
CUSTOM = "custom"
TITLE = "title"
PAYLOAD = "payload"

# Used for LLM command generation
ROLE_USER = "user"
ROLE_SYSTEM = "system"

# Used for key values in ValidateSlotPatternFlowStackFrame
REFILL_UTTER = "refill_utter"
REJECTIONS = "rejections"
